# Licensed under a 3-clause BSD style license - see LICENSE.rst
"""This module defines thermal spectra."""

# ASTROPY
from astropy import units as u
from astropy.io import fits
from astropy.io.fits.connect import is_fits

# LOCAL
from synphot import exceptions, specio, units
from synphot.models import BlackBody1D, Empirical1D
from synphot.spectrum import BaseUnitlessSpectrum, SourceSpectrum

__all__ = ['ThermalSpectralElement']


class ThermalSpectralElement(BaseUnitlessSpectrum):
    """Class to handle spectral element with associated thermal
    properties.

    This differs from `~synphot.spectrum.SpectralElement` in the
    sense that it carries thermal parameters, i.e., temperature
    and beam filling factor.

    .. note::

        Use :func:`thermal_source` to apply its emissivity to an
        existing beam.

    Parameters
    ----------
    modelclass, kwargs
        See `~synphot.spectrum.BaseSpectrum`.

    temperature : float or `~astropy.units.quantity.Quantity`
        Temperature. If not a Quantity, assumed to be in Kelvin.

    beam_fill_factor : float or `~astropy.units.quantity.Quantity`
        Beam filling factor. If a Quantity, must be unitless.
        Defaults to 1.

    """
    def __init__(self, modelclass, temperature, beam_fill_factor=1, **kwargs):
        super(ThermalSpectralElement, self).__init__(modelclass, **kwargs)
        self.temperature = temperature
        self.beam_fill_factor = beam_fill_factor

    @property
    def temperature(self):
        """Temperature."""
        return self._temperature

    @temperature.setter
    def temperature(self, what):
        """Set temperature."""
        self._temperature = units.validate_quantity(what, u.K)

    @property
    def beam_fill_factor(self):
        """Beam filling factor."""
        return self._beam_fill_factor

    @beam_fill_factor.setter
    def beam_fill_factor(self, what):
        """Set beam filling factor."""
        self._beam_fill_factor = units.validate_quantity(what, '').value

    def taper(self, **kwargs):
        """Tapering is disabled."""
        raise NotImplementedError(
            'Thermal spectral element cannot be tapered.')

    def thermal_source(self):
        """Apply emissivity to an existing beam to produce a thermal
        source spectrum (without optical counterpart).

        Thermal source spectrum is calculated as follow:

            #. Create a blackbody spectrum in PHOTLAM per square arcsec
               with `temperature`.
            #. Multiply the blackbody with `beam_fill_factor` and ``self``.

        Returns
        -------
        sp : `~synphot.spectrum.SourceSpectrum`
            Thermal source spectrum.

        """
        sp = (SourceSpectrum(BlackBody1D, temperature=self.temperature) *
              units.SR_PER_ARCSEC2 * self.beam_fill_factor * self)
        sp.meta['temperature'] = self.temperature
        sp.meta['beam_fill_factor'] = self.beam_fill_factor
        return sp

    @classmethod
    def from_file(cls, filename, temperature_key='DEFT',
                  beamfill_key='BEAMFILL', **kwargs):
        """Creates a thermal spectral element from file.

        .. note::

            Only FITS format is supported.

        Parameters
        ----------
        filename : str
            Thermal spectral element filename.

        temperature_key, beamfill_key : str
            Keywords in FITS *table extension* that store temperature
            (in Kelvin) and beam filling factor values.
            Beam filling factor is set to 1 if its keyword is missing.

        kwargs : dict
            Keywords acceptable by :func:`~synphot.specio.read_fits_spec`.

        Returns
        -------
        th : `ThermalSpectralElement`
            Empirical thermal spectral element.

        Raises
        ------
        synphot.exceptions.SynphotError
            Invalid inputs.

        """
        if not is_fits("", filename, None):
            raise exceptions.SynphotError('Only FITS format is supported.')

        # Extra info from table header
        ext = kwargs.get('ext', 1)
        tab_hdr = fits.getheader(filename, ext=ext)

        temperature = tab_hdr.get(temperature_key)
        if temperature is None:
            raise exceptions.SynphotError(
                'Missing {0} keyword.'.format(temperature_key))

        beam_fill_factor = tab_hdr.get('BEAMFILL', 1)

        if 'flux_col' not in kwargs:
            kwargs['flux_col'] = 'EMISSIVITY'

        header, wavelengths, em = specio.read_spec(filename, **kwargs)
        return cls(
            Empirical1D, temperature, beam_fill_factor=beam_fill_factor,
            points=wavelengths, lookup_table=em, meta={'header': header})



"""This should be updated if blackbody.py is removed from synphot."""

import numpy as np
from astropy import constants as const
from astropy import units as u

from synphot.blackbody import blackbody_nu
from synphot.units import FNU


# This test was removed from astropy in
# https://github.com/astropy/astropy/pull/9282
# but the rest of blackbody.py tests are still over there on astropy.
def test_blackbody_synphot():
    """Test that it is consistent with IRAF SYNPHOT BBFUNC."""
    # Solid angle of solar radius at 1 kpc
    fac = np.pi * (const.R_sun / const.kpc) ** 2 * u.sr

    with np.errstate(all='ignore'):
        flux = blackbody_nu([100, 1, 1000, 1e4, 1e5] * u.AA, 5000) * fac
    assert flux.unit == FNU

    # Special check for overflow value (SYNPHOT gives 0)
    assert np.log10(flux[0].value) < -143

    np.testing.assert_allclose(
        flux.value[1:], [0, 2.01950807e-34, 3.78584515e-26, 1.90431881e-27],
        rtol=0.01)  # 1% accuracy



# -*- coding: utf-8 -*-
# Licensed under a 3-clause BSD style license - see LICENSE.rst
#
# Astropy documentation build configuration file.
#
# This file is execfile()d with the current directory set to its containing
# dir.
#
# Note that not all possible configuration values are present in this file.
#
# All configuration values have a default. Some values are defined in
# the global Astropy configuration which is loaded here before anything else.
# See sphinx_astropy.conf for which values are set there.

import datetime
import os
import sys
from pkg_resources import get_distribution

# Load all of the global Astropy configuration
try:
    from sphinx_astropy.conf.v1 import *  # noqa
except ImportError:
    print('ERROR: the documentation requires the sphinx-astropy package '
          'to be installed')
    sys.exit(1)

# Get configuration information from setup.cfg
from configparser import ConfigParser
conf = ConfigParser()
conf.read([os.path.join(os.path.dirname(__file__), '..', 'setup.cfg')])
setup_cfg = dict(conf.items('metadata'))

# -- General configuration ----------------------------------------------------

# If your documentation needs a minimal Sphinx version, state it here.
# needs_sphinx = '1.2'

# To perform a Sphinx version check that needs to be more specific than
# major.minor, call `check_sphinx_version("x.y.z")` here.
# check_sphinx_version("1.2.1")

# List of patterns, relative to source directory, that match files and
# directories to ignore when looking for source files.
exclude_patterns.append('_templates')

# This is added to the end of RST files - a good place to put substitutions to
# be used globally.
rst_epilog += """
"""

# -- Project information ------------------------------------------------------

# This does not *have* to match the package name, but typically does
project = setup_cfg['name']
author = setup_cfg['author']
copyright = '{0}, {1}'.format(datetime.datetime.now().year, author)

# The version info for the project you're documenting, acts as replacement for
# |version| and |release|, also used in various other places throughout the
# built documents.
#
# The full version, including alpha/beta/rc tags.
release = get_distribution(project).version
# The short X.Y version.
version = '.'.join(release.split('.')[:2])

# A list of ignored prefixes for module index sorting.
modindex_common_prefix = ['synphot.']

# -- Options for HTML output --------------------------------------------------

# A NOTE ON HTML THEMES
# The global astropy configuration uses a custom theme, 'bootstrap-astropy',
# which is installed along with astropy. A different theme can be used or
# the options for this theme can be modified by overriding some of the
# variables set in the global configuration. The variables set in the
# global configuration are listed below, commented out.

# Add any paths that contain custom themes here, relative to this directory.
# To use a different custom theme, add the directory containing the theme.
# html_theme_path = []

# The theme to use for HTML and HTML Help pages.  See the documentation for
# a list of builtin themes. To override the custom theme, set this to the
# name of a builtin theme or the name of a custom theme in html_theme_path.
# html_theme = 'sphinx_rtd_theme'

# Custom sidebar templates, maps document names to template names.
# html_sidebars = {}

# Logo
html_logo = '_static/stsci_logo.png'

# The name of an image file (within the static path) to use as favicon of the
# docs.  This file should be a Windows icon file (.ico) being 16x16 or 32x32
# pixels large.
# html_favicon = ''

# If not '', a 'Last updated on:' timestamp is inserted at every page bottom,
# using the given strftime format.
# html_last_updated_fmt = ''

# The name for this set of Sphinx documents.  If None, it defaults to
# "<project> v<release> documentation".
html_title = '{0} v{1}'.format(project, release)

# Output file base name for HTML help builder.
htmlhelp_basename = project + 'doc'


# -- Options for LaTeX output -------------------------------------------------

# Grouping the document tree into LaTeX files. List of tuples
# (source start file, target name, title, author, documentclass [howto/manual])
latex_documents = [('index', project + '.tex', project + u' Documentation',
                    author, 'manual')]


# -- Options for manual page output -------------------------------------------

# One entry per manual page. List of tuples
# (source start file, name, description, authors, manual section).
man_pages = [('index', project.lower(), project + u' Documentation',
              [author], 1)]


# -- Options for intersphinx --------------------------------------------------

intersphinx_mapping.update({
    'astropy': ('https://docs.astropy.org/en/stable/', None),
    'stsynphot': ('https://stsynphot.readthedocs.io/en/latest/', None),
    'specutils': ('https://specutils.readthedocs.io/en/stable/', None),
    'dust-extinction': ('https://dust-extinction.readthedocs.io/en/stable/', None)})

# -- Options for linkcheck output ---------------------------------------------
linkcheck_retry = 5
linkcheck_ignore = ['https://hsthelp.stsci.edu']
linkcheck_timeout = 180
linkcheck_anchors = False

# Enable nitpicky mode - which ensures that all references in the docs resolve.
nitpicky = True
nitpick_ignore = [
    ('py:class', 'synphot.models.BaseGaussian1D'),
]



"""Handle Fast Fourier Transform (FFT) for filter parameterization."""

import numpy as np
from astropy import units as u
from astropy.modeling.models import custom_model, Sine1D
from astropy.table import Table

from synphot.models import Empirical1D
from synphot.spectrum import SpectralElement
from synphot.units import validate_quantity

__all__ = ['filter_to_fft', 'filter_from_fft', 'analytical_model_from_fft',
           'filters_to_fft_table']


def _simplified_wavelength(n_lambda, lambda_0, delta_lambda):
    # tynt assumed everything was in Angstrom, which coincides with
    # synphot internal wavelength unit.
    wave_unit = SpectralElement._internal_wave_unit

    lambda_0 = validate_quantity(
        lambda_0, wave_unit, equivalencies=u.spectral())
    delta_lambda = validate_quantity(
        delta_lambda, wave_unit, equivalencies=u.spectral())
    lambda_max = (n_lambda + 1) * delta_lambda + lambda_0

    return np.arange(lambda_0.value, lambda_max.value,
                     delta_lambda.value) * wave_unit


def filter_to_fft(bp, wavelengths=None, n_terms=10):
    """Calculate filter parameters using FFT.

    Parameters
    ----------
    bp : `~synphot.spectrum.SpectralElement`
        Filter to parameterize.

    wavelengths : array-like or `~astropy.units.quantity.Quantity`
        Wavelength values for sampling.
        If not a Quantity, assumed to be in Angstrom.
        If `None`, ``waveset`` is used.

    n_terms : int
        Number of FFT parameters to keep.

    Returns
    -------
    n_lambda : int
        Number of elements in ``wl``.

    lambda_0 : `~astropy.units.quantity.Quantity`
        Minimum value of ``wl``.

    delta_lambda : `~astropy.units.quantity.Quantity`
        Median delta wavelength.

    tr_max : `~astropy.units.quantity.Quantity`
        Maximum value of ``tr``.

    fft_parameters : list of complex
        List of complex values that are FFT parameters to keep.

    """
    wl = bp._validate_wavelengths(wavelengths)
    tr = bp(wl)

    diff_wl = np.diff(wl)

    delta_lambda = np.nanmedian(diff_wl[diff_wl != 0])
    lambda_0 = wl.min()
    n_lambda = len(wl)

    # Create a simplified wavelength grid
    simplified_wavelength = _simplified_wavelength(
        n_lambda, lambda_0, delta_lambda)

    tr_max = tr.max()

    # Interpolate transmittance onto simplified wavelength grid
    tr_interp = np.interp(simplified_wavelength, wl, tr)

    # Take the DFT of the interpolated transmittance curve
    fft = np.fft.fft(tr_interp)[:n_terms]

    if isinstance(fft, u.Quantity):
        fft_parameters = fft.value.tolist()
    else:  # Older Numpy does not return Quantity
        fft_parameters = fft.tolist()

    return n_lambda, lambda_0, delta_lambda, tr_max, fft_parameters


def filter_from_fft(n_lambda, lambda_0, delta_lambda, tr_max, fft_parameters):
    """Reconstruct a filter from given FFT parameters.
    The inputs for this function can be obtained from :func:`filter_to_fft`.

    Parameters
    ----------
    n_lambda : int
        Number of elements in original wavelength array.

    lambda_0 : float or `~astropy.units.quantity.Quantity`
        Minimum value of original wavelength array.
        If not a Quantity, assumed to be in Angstrom.

    delta_lambda : float or `~astropy.units.quantity.Quantity`
        Median delta wavelength of original wavelength array.
        If not a Quantity, assumed to be in Angstrom.

    tr_max : float or `~astropy.units.quantity.Quantity`
        Maximum value of transmittance curve.
        If a Quantity, must be unitless.

    fft_parameters : list of complex
        List of complex values that are FFT parameters representing the
        filter transmittance curve.

    Returns
    -------
    bp : `~synphot.spectrum.SpectralElement`
       Reconstructed filter.

    """
    wavelength = _simplified_wavelength(n_lambda, lambda_0, delta_lambda)
    n_wave = len(wavelength)
    ifft = np.fft.ifft(fft_parameters, n=n_wave)
    transmittance = ((ifft.real - ifft.real.min()) * tr_max / np.ptp(ifft.real))  # noqa
    return SpectralElement(
        Empirical1D, points=wavelength, lookup_table=transmittance)


def analytical_model_from_fft(n_lambda, lambda_0, delta_lambda, tr_max,
                              fft_parameters):
    """Similar to :func:`filter_from_fft` except that this returns
    an analytical model.

    .. note::

        This model needs to be sampled using the full range of
        wavelength. See https://github.com/bmorris3/tynt/issues/9 .

    Returns
    -------
    astropy_model : `~astropy.modeling.CompoundModel`
        A compound model that consists of
        `~astropy.modeling.functional_models.Sine1D` models.

    """
    wavelength = _simplified_wavelength(n_lambda, lambda_0, delta_lambda)
    n_wave = len(wavelength)
    n_fft_pars = len(fft_parameters)

    m = (np.sum([Sine1D(amplitude=fft_parameters[i].real / n_wave,
                        frequency=i / n_wave, phase=0.25)
                 for i in range(n_fft_pars)]) -
         np.sum([Sine1D(amplitude=fft_parameters[i].imag / n_wave,
                        frequency=i / n_wave)
                 for i in range(n_fft_pars)]))

    @custom_model
    def fft_model(x):
        """Approximate Fourier reconstruction of an astronomical filter.

        Parameters
        ----------
        x : `~astropy.units.quantity.Quantity`
            Full wavelength range that samples the filter.

        Returns
        -------
        transmittance : array-like or `~astropy.units.quantity.Quantity`
            Transmittance curve. If ``tr_max`` is a Quantity, this will
            be a Quantity as well.

        """
        wave_unit = SpectralElement._internal_wave_unit
        x = validate_quantity(x, wave_unit, equivalencies=u.spectral())

        mo = m((x - wavelength.min()) / (wavelength[1] - wavelength[0]))
        return (mo - mo.min()) * tr_max / np.ptp(mo)

    return fft_model()


def filters_to_fft_table(filters_mapping, n_terms=10):
    """Run :func:`filter_to_fft` on a list of filters
    and store results in a table.

    Parameters
    ----------
    filters_mapping : dict
        Dictionary mapping human-readable filter name to its
        `~synphot.spectrum.SpectralElement` and wavelengths, if applicable.
        If the filter object has a valid ``waveset``, just provide `None`
        for wavelengths; otherwise provide a Quantity array for sampling.
        For example::

            {'JOHNSON/V': (<SpectralElement ...>, None),
             'Flat': (<SpectralElement ...>, <Quantity [1000., ..., 9999.] Angstrom>)}

    n_terms : int
        Number of FFT parameters to keep.

    Returns
    -------
    fft_table : `~astropy.table.Table`
        Table storing FFT parameterization for the given filters.
        Use its ``write`` method to save it to file.

    """  # noqa
    wave_unit = SpectralElement._internal_wave_unit
    colnames = ['filter', 'n_lambda', 'lambda_0', 'delta_lambda',
                'tr_max'] + [f'fft_{i}' for i in range(n_terms)]
    rows = []

    for key, (bp, wavelengths) in filters_mapping.items():
        n_lambda, lambda_0, delta_lambda, tr_max, fft_pars = filter_to_fft(
            bp, wavelengths=wavelengths, n_terms=n_terms)
        rows.append(tuple(
            [key, n_lambda, lambda_0, delta_lambda, tr_max] + fft_pars))

    fft_table = Table(rows=rows, names=colnames)
    fft_table['lambda_0'].unit = wave_unit
    fft_table['delta_lambda'].unit = wave_unit

    return fft_table



# Licensed under a 3-clause BSD style license - see LICENSE.rst
"""Test observation.py module."""

# STDLIB
import os
import warnings

# THIRD-PARTY
import numpy as np
import pytest

# ASTROPY
from astropy import units as u
from astropy.modeling.models import Const1D
from astropy.tests.helper import assert_quantity_allclose
from astropy.utils.data import get_pkg_data_filename
from astropy.utils.exceptions import (AstropyDeprecationWarning,
                                      AstropyUserWarning)

# LOCAL
from synphot.tests.test_units import _area
from synphot import exceptions, units
from synphot.compat import HAS_SPECUTILS  # noqa
from synphot.models import (
    BlackBodyNorm1D, Box1D, ConstFlux1D, Empirical1D, GaussianFlux1D
)
from synphot.observation import Observation
from synphot.spectrum import SourceSpectrum, SpectralElement

# Global test data files
_specfile = get_pkg_data_filename(
    os.path.join('data', 'grw_70d5824_stisnic_005.fits'),
    package='synphot.tests'
)
_bandfile = get_pkg_data_filename(
    os.path.join('data', 'hst_acs_hrc_f555w.fits'),
    package='synphot.tests'
)


class TestObservation:
    """Test Observation (most of them)."""
    def setup_class(self):
        sp = SourceSpectrum(
            ConstFlux1D, amplitude=1 * units.FLAM,
            meta={'warnings': {'w1': 'spec warn', 'w2': 'foo'}})
        bp = SpectralElement.from_file(_bandfile)
        bp.warnings = {'w1': 'band warn'}
        w = np.arange(1000, 11001, dtype=np.float64)
        self.obs = Observation(sp, bp, binset=w)

    def test_invalid_input_spec(self):
        with pytest.raises(exceptions.SynphotError):
            Observation(self.obs.bandpass, self.obs.bandpass)
        with pytest.raises(exceptions.SynphotError):
            Observation(self.obs.spectrum, self.obs.spectrum)

    def test_inherit_warnings(self):
        assert sorted(self.obs.warnings) == ['w1', 'w2']
        assert self.obs.warnings['w1'] == 'band warn'

    def test_disjoint_spec(self):
        """The rest of overlap logic is tested in `TestInitWithForce`."""
        sp = SourceSpectrum(
            Empirical1D, points=[39999.9, 40000, 40001, 40001.1],
            lookup_table=[0, 1, 1, 0])
        with pytest.raises(exceptions.DisjointError):
            Observation(sp, self.obs.bandpass)

    def test_taper(self):
        with pytest.raises(NotImplementedError):
            self.obs.taper()

    def test_binned_data(self):
        # Binned flux
        np.testing.assert_array_equal(self.obs.binflux[:10], 0)
        np.testing.assert_array_equal(self.obs.binflux[-10:], 0)
        np.testing.assert_allclose(
            self.obs.sample_binned(wavelengths=self.obs.binset[5000:5010],
                                   flux_unit=units.FLAM).value,
            [0.12265425, 0.12226972, 0.12184207, 0.12141429, 0.12098646,
             0.1205586, 0.1201307, 0.11970269, 0.11927488, 0.11884699],
            rtol=1e-4)

        # Binned wave and flux
        w, y = self.obs._get_binned_arrays(
            [599.9999999999, 600.4, 600.9] * u.nm, units.FLAM)
        np.testing.assert_allclose(w.value, [600, 600.4, 600.9])
        np.testing.assert_allclose(
            y.value, [0.12265425, 0.12098646, 0.11884699], rtol=1e-4)

        w, y = self.obs._get_binned_arrays(None, units.PHOTLAM)
        np.testing.assert_array_equal(w, self.obs.binset)
        np.testing.assert_array_equal(y, self.obs.binflux)

        # Bin edges
        np.testing.assert_array_equal(
            self.obs.bin_edges.value[:10],
            [999.5, 1000.5, 1001.5, 1002.5, 1003.5, 1004.5, 1005.5, 1006.5,
             1007.5, 1008.5])
        np.testing.assert_array_equal(
            self.obs.bin_edges.value[5000:5010],
            [5999.5, 6000.5, 6001.5, 6002.5, 6003.5, 6004.5, 6005.5, 6006.5,
             6007.5, 6008.5])
        np.testing.assert_array_equal(
            self.obs.bin_edges.value[-10:],
            [10991.5, 10992.5, 10993.5, 10994.5, 10995.5, 10996.5, 10997.5,
             10998.5, 10999.5, 11000.5])

    def test_reversed_binset(self):
        obs2 = Observation(
            self.obs.spectrum, self.obs.bandpass, binset=self.obs.binset[::-1])
        np.testing.assert_array_equal(obs2.binset, self.obs.binset)
        np.testing.assert_array_equal(obs2.binflux, self.obs.binflux)
        np.testing.assert_array_equal(obs2.bin_edges, self.obs.bin_edges)

    def test_sampled_binned_exceptions(self):
        with pytest.raises(exceptions.InterpolationNotAllowed):
            self.obs.sample_binned([6000, 6004.5, 6009])
        with pytest.raises(exceptions.UnsortedWavelength):
            self.obs.sample_binned([6004, 6000, 6009])

    @pytest.mark.parametrize(
        ('cenwave', 'ans'),
        [(500 * u.nm, [499.9, 500.1] * u.nm),
         (5000, [4999, 5001] * u.AA)])
    def test_binned_waverange(self, cenwave, ans):
        w = self.obs.binned_waverange(cenwave, 2, mode='none')
        np.testing.assert_allclose(w, ans)

    def test_binned_pixelrange(self):
        w = [499.95, 500.05] * u.nm
        assert self.obs.binned_pixelrange(w, mode='round') == 1

    def test_default_binset_from_bandpass(self):
        obs2 = Observation(self.obs.spectrum, self.obs.bandpass)
        np.testing.assert_array_equal(obs2.binset, self.obs.bandpass.waveset)

    def test_default_binset_from_spectrum(self):
        tf_unit = u.erg / (u.cm * u.cm * u.s)
        sp = SourceSpectrum(
            GaussianFlux1D, mean=5000, total_flux=(1 * tf_unit), fwhm=10)
        bp = SpectralElement(Const1D, amplitude=1)
        with pytest.warns(AstropyUserWarning,
                          match=r'Source spectrum will be evaluated outside '
                          r'pre-defined waveset'):
            obs2 = Observation(sp, bp, force='extrap')
        np.testing.assert_array_equal(obs2.binset, sp.waveset)

    def test_undefined_binset(self):
        bp = SpectralElement(Const1D, amplitude=1)
        with pytest.raises(exceptions.UndefinedBinset):
            Observation(self.obs.spectrum, bp)

    def test_as_spectrum(self):
        w = np.arange(6000, 6005)
        sp1 = self.obs.as_spectrum(binned=True, wavelengths=w)  # Binned
        sp2 = self.obs.as_spectrum(binned=False, wavelengths=w)  # Sampled
        np.testing.assert_allclose(
            sp1(sp1.waveset), sp2(sp2.waveset), rtol=1e-3)


class TestInitWithForce:
    """Test forced initialization."""
    def setup_class(self):
        x = np.arange(3000, 4000)
        y = np.ones_like(x) * 0.75
        self.sp = SourceSpectrum(
            Empirical1D, points=x, lookup_table=y, meta={'expr': 'short flat'})
        self.bp = SpectralElement(
            Empirical1D, points=[3949.9, 3950, 4050, 4050.1],
            lookup_table=[0, 1, 1, 0])

    @pytest.mark.parametrize(
        ('force_type', 'ans'),
        [('taper', 0),
         ('extrap', 0.75)])
    def test_force(self, force_type, ans):
        with warnings.catch_warnings():
            warnings.filterwarnings(
                'ignore', message=r'.*Source spectrum.*',
                category=AstropyUserWarning)
            obs = Observation(self.sp, self.bp, force=force_type)
        assert obs(4005).value == ans
        assert force_type in obs.warnings['PartialOverlap']

    def test_exceptions(self):
        with pytest.raises(exceptions.PartialOverlap):
            Observation(self.sp, self.bp)
        with pytest.raises(exceptions.SynphotError):
            Observation(self.sp, self.bp, force='foo')


class TestMathOperators:
    """Test Observation math operators."""
    def setup_class(self):
        sp = SourceSpectrum(ConstFlux1D, amplitude=1)
        bp = SpectralElement(Box1D, amplitude=1, x_0=5000, width=100)
        w = np.arange(1000, 10000)
        self.obs = Observation(sp, bp, binset=w)

    @pytest.mark.parametrize('is_scalar', [True, False])
    def test_mul(self, is_scalar):
        if is_scalar:
            other = 2
        else:
            other = SpectralElement(Const1D, amplitude=2)

        obs2 = self.obs * other
        np.testing.assert_allclose(obs2([1000, 5000]).value, [0, 2])
        np.testing.assert_allclose(
            obs2.sample_binned([4000, 4999]).value, [0, 2])

    def test_div(self):
        with pytest.raises(NotImplementedError):
            self.obs / 2

    def test_addsub(self):
        with pytest.raises(NotImplementedError):
            self.obs + self.obs
        with pytest.raises(NotImplementedError):
            self.obs - self.obs


class TestObsPar:
    """Test Observation values from IRAF SYNPHOT CALCPHOT,
    unless noted otherwise.

    """
    def setup_class(self):
        sp = SourceSpectrum.from_file(_specfile)
        bp = SpectralElement.from_file(_bandfile)
        self.obs = Observation(sp, bp)

    def test_avglam(self):
        """Tested for PHOTLAM only; no binning."""
        np.testing.assert_allclose(
            self.obs.avgwave().value, 5321.091, rtol=1e-5)

    def test_barlam(self):
        """Tested for PHOTLAM only; no binning."""
        np.testing.assert_allclose(
            self.obs.barlam().value, 5286.685, rtol=1e-5)

    def test_pivot(self):
        """Tested with value from ASTROLIB PYSYNPHOT; no binning."""
        np.testing.assert_allclose(self.obs.pivot().value, 5309.578, rtol=1e-5)

    def test_normalize(self):
        """Tested with value from ASTROLIB PYSYNPHOT; no binning."""
        obs2 = self.obs.normalize(1 * units.PHOTLAM, band=self.obs.bandpass)
        np.testing.assert_allclose(
            obs2.countrate(_area).value, 59517756.384, rtol=1e-5)

    @pytest.mark.parametrize('binned', [True, False])
    def test_efflam(self, binned):
        with pytest.warns(AstropyDeprecationWarning,
                          match='Usage of EFFLPHOT is deprecated') as w:
            x = self.obs.effective_wavelength(mode='efflphot', binned=binned)
        assert len(w) == 1

        np.testing.assert_allclose(x.value, 5344.312, rtol=1e-4)

    @pytest.mark.parametrize('binned', [True, False])
    def test_efflam_erg(self, binned):
        np.testing.assert_allclose(
            self.obs.effective_wavelength(mode='efflerg', binned=binned).value,
            5321.091, rtol=1e-4)

    def test_efflam_exception(self):
        with pytest.raises(exceptions.SynphotError):
            self.obs.effective_wavelength(mode='foo')

    # ans taken from calculating effstim in FLAM and then using unit
    # conversion around bandpass pivot wavelength.
    @pytest.mark.parametrize(
        ('flux_unit', 'ans'),
        [(units.FLAM, 3.05131543e-14),
         (units.FNU, 2.91961387e-25),
         (None, 0.00822697),
         (units.PHOTLAM, 0.00822697),
         (units.PHOTNU, 7.8718752e-14),
         (u.Jy, 0.02919614),
         (u.mJy, 29.19613866)])
    def test_effstim(self, flux_unit, ans):
        """Test EFFSTIM for all supported non-mag and non-count flux units."""
        np.testing.assert_allclose(
            self.obs.effstim(flux_unit=flux_unit).value, ans)

    def test_effstim_count(self):
        """Test EFFSTIM in count and OBMAG separately due to different API
        and tolerance.

        .. note::

            ``binned``, ``waverange``, and ``force`` tested in `TestCountRate`.

        """
        # pysynphot 0.9.12.dev5
        ans_ct = 101462.39601864747
        ans_ob = -12.515762784747238
        tol = 1e-4

        val_ct = self.obs.effstim(flux_unit=u.count, area=_area).value
        np.testing.assert_allclose(val_ct, ans_ct, rtol=tol)

        val_ob = self.obs.effstim(flux_unit=units.OBMAG, area=_area).value
        np.testing.assert_allclose(val_ob, ans_ob, atol=tol, rtol=0)

        # Sanity check
        np.testing.assert_allclose(val_ob, -2.5 * np.log10(val_ct))

    # ans taken from calculating effstim in FLAM and then using unit
    # conversion around bandpass pivot wavelength.
    @pytest.mark.parametrize(
        ('flux_unit', 'ans'),
        [(u.STmag, 12.68878224),
         (u.ABmag, 12.73668646)])
    def test_effstim_mag(self, flux_unit, ans):
        """Test mag separately because tolerance calculation, if needed,
        is different."""
        np.testing.assert_allclose(
            self.obs.effstim(flux_unit=flux_unit, area=_area).value, ans)

    def test_effstim_analytic(self):
        sp = SourceSpectrum(BlackBodyNorm1D, temperature=5000)
        bp = SpectralElement(Box1D, amplitude=1, x_0=5500, width=1)
        obs = Observation(sp, bp)
        np.testing.assert_allclose(
            obs.effstim(flux_unit=units.FLAM).value, 2.03E-15, rtol=0.01)  # 1%

    @pytest.mark.remote_data
    def test_effstim_vegamag(self):
        ans = 12.737293324241517  # pysynphot 0.9.12.dev5
        vspec = SourceSpectrum.from_vega()
        np.testing.assert_allclose(
            self.obs.effstim(flux_unit=units.VEGAMAG, vegaspec=vspec).value,
            ans, rtol=0.001)

    @pytest.mark.parametrize('flux_unit', [u.mag, units.VEGAMAG])
    def test_effstim_exceptions(self, flux_unit):
        with pytest.raises(exceptions.SynphotError):
            self.obs.effstim(flux_unit=flux_unit)
        with pytest.raises(exceptions.SynphotError):
            self.obs.effstim(units.VEGAMAG)


class TestCountRate:
    """Test countrate with Observation with well-defined ranges.

    .. note:: Use binned data except for :func:`test_waverange_no_bin`.

    """
    def setup_class(self):
        x = np.arange(1000, 1100, 0.5) * u.AA
        y = units.convert_flux(
            x, (x.value - 1000) * u.count, units.PHOTLAM, area=_area)
        self.sp = SourceSpectrum(Empirical1D, points=x, lookup_table=y,
                                 meta={'expr': 'slope1'})
        self.bp = SpectralElement(
            Empirical1D, points=[1009.95, 1010, 1030, 1030.05] * u.AA,
            lookup_table=[0, 1, 1, 0], meta={'expr': 'handmade_box'})
        self.binset = np.arange(1000, 1020) * u.AA
        self.obs = Observation(self.sp, self.bp, binset=self.binset)

    @pytest.mark.skipif('not HAS_SPECUTILS')
    def test_spectrum1d_source(self):
        # Should round-trip. See also: test_waverange(w=None)
        spec = self.sp.to_spectrum1d()
        obs = Observation(spec, self.bp, binset=self.binset)
        assert_quantity_allclose(
            obs.countrate(_area), 280.75 * (u.count / u.s))

    @pytest.mark.parametrize(
        ('w', 'ans'),
        [(None, 280.75 * (u.count / u.s)),
         ([1000, 1019] * u.AA, 280.75 * (u.count / u.s)),
         ([1013, 1016] * u.AA, 116 * (u.count / u.s)),
         ([1012.8, 1016] * u.AA, 116 * (u.count / u.s)),
         ([1013.2, 1016] * u.AA, 116 * (u.count / u.s))])
    def test_waverange(self, w, ans):
        """Use given wavelength range on binned data."""
        assert_quantity_allclose(self.obs.countrate(_area, waverange=w), ans)

    def test_waverange_no_bin(self):
        """Use given wavelength range on native dataset.

        .. note:: Accuracy unsure as there was no precedent to test against.

        """
        w = [1000, 1019]
        np.testing.assert_allclose(
            self.obs.countrate(_area, waverange=w, binned=False).value, 271)

    @pytest.mark.parametrize(
        ('w', 'ans'),
        [([1016, 1026], 140),
         ([999, 1016], 172.75)])
    def test_force(self, w, ans):
        """Force calculation for partial overlap."""
        with pytest.warns(AstropyUserWarning, match=r'Count rate calculated '
                          r'only for wavelengths in the overlap between '
                          r'observation and given range'):
            np.testing.assert_allclose(
                self.obs.countrate(_area, waverange=w, force=True).value, ans)

        # Must raise error without force
        with pytest.raises(exceptions.PartialOverlap):
            self.obs.countrate(_area, waverange=w)

    def test_disjoint_waverange(self):
        with pytest.raises(exceptions.DisjointError):
            self.obs.countrate(_area, waverange=[1020, 1030])


class TestCountRateNegFlux:
    """Test countrate with files containing negative flux/throughput values."""
    def setup_class(self):
        self.bp = SpectralElement.from_file(get_pkg_data_filename(
            os.path.join('data', 'cos_fuv_g130m_c1309_psa.fits'),
            package='synphot.tests'))
        self.spfile = get_pkg_data_filename(
            os.path.join('data', 'us7.txt'), package='synphot.tests'
        )

    @pytest.mark.parametrize(
        ('keep_neg', 'ans'),
        [(True, 1510.219531414891),
         (False, 1627.8250215634343)])
    def test_neg_handling(self, keep_neg, ans):
        with warnings.catch_warnings():
            warnings.filterwarnings(
                'ignore', message=r'.*contained negative flux or throughput.*',
                category=AstropyUserWarning)
            sp = SourceSpectrum.from_file(self.spfile, keep_neg=keep_neg)
        obs = Observation(sp, self.bp)
        c = obs.countrate(_area)
        np.testing.assert_allclose(c.value, ans, rtol=1e-4)

        if not keep_neg:
            assert 'NegativeFlux' in obs.warnings


def test_countrate_neg_leak():
    """Test countrate of sub-sampling not exceeding total countrate.
    https://github.com/spacetelescope/synphot_refactor/issues/126
    """
    # This bug only manifests itself in very specific cases.
    bp = SpectralElement.from_file(get_pkg_data_filename(
        os.path.join('data', 'stis_fuv_f25ndq2_mjd58300_0822774.fits'),
        package='synphot.tests'))
    sp = SourceSpectrum.from_file(get_pkg_data_filename(
        os.path.join('data', 'k93_4500_0_5_rn_box.fits'),
        package='synphot.tests'))
    binset = np.fromfile(get_pkg_data_filename(
        os.path.join('data', 'stis_fuv_f25ndq2_binset.bin'),
        package='synphot.tests'), dtype='<f8')
    obs = Observation(sp, bp, binset=binset)
    area = 45238.93416  # HST cm^2
    wrange = [1109.22, 12000.0]  # Angstrom
    c_all = obs.countrate(area)
    c_sub = obs.countrate(area, waverange=wrange)
    assert c_sub <= c_all



try:
    from pytest_astropy_header.display import (PYTEST_HEADER_MODULES,
                                               TESTED_VERSIONS)
except ImportError:
    PYTEST_HEADER_MODULES = {}
    TESTED_VERSIONS = {}

try:
    from synphot import __version__ as version
except ImportError:
    version = 'unknown'

from astropy.utils import minversion
NUMPY_LT_2_0 = not minversion("numpy", "2.0.dev")
if not NUMPY_LT_2_0:
    import numpy as np
    np.set_printoptions(legacy="1.25")

# Uncomment and customize the following lines to add/remove entries
# from the list of packages for which version numbers are displayed
# when running the tests.
PYTEST_HEADER_MODULES['astropy'] = 'astropy'
PYTEST_HEADER_MODULES['specutils'] = 'specutils'
PYTEST_HEADER_MODULES['dust-extinction'] = 'dust_extinction'
PYTEST_HEADER_MODULES.pop('Matplotlib', None)
PYTEST_HEADER_MODULES.pop('Pandas', None)
PYTEST_HEADER_MODULES.pop('h5py', None)

TESTED_VERSIONS['synphot'] = version



# Licensed under a 3-clause BSD style license - see LICENSE.rst
"""This module defines an observed spectrum, i.e., a source spectrum
that has gone through a bandpass.

"""

# STDLIB
import math
import warnings

# THIRD-PARTY
import numpy as np
from scipy.integrate import trapezoid

# ASTROPY
from astropy import log
from astropy import units as u
from astropy.utils.exceptions import (AstropyUserWarning,
                                      AstropyDeprecationWarning)

# LOCAL
from synphot import binning, exceptions, units, utils
from synphot.models import Empirical1D
from synphot.spectrum import (BaseSourceSpectrum, SourceSpectrum,
                              SpectralElement)

__all__ = ['Observation']


class Observation(BaseSourceSpectrum):
    """This is an observed spectrum, where a source spectrum
    has gone through a bandpass.

    Usually, this is the end point of a chain of spectral
    manipulation. It has extra attributes that deal with binning,
    which is introduced by the detector.

    Parameters
    ----------
    spec : `~synphot.spectrum.SourceSpectrum` or `specutils.Spectrum1D`
        Source spectrum.

    band : `~synphot.spectrum.SpectralElement`
        Bandpass.

    binset : array-like, `~astropy.units.quantity.Quantity`, or `None`
        Center of binned wavelengths.
        If not a Quantity, assumed to be in Angstrom.
        If `None`, input ``self.waveset`` values are used.

    force : {`None`, 'none', 'extrap', 'taper'}
        Force creation of an observation even when source spectrum
        and bandpass do not fully overlap:

            * `None` or 'none' - Source must encompass bandpass (default)
            * 'extrap' - Extrapolate source spectrum (this changes the
              underlying model of ``spec`` to always extrapolate,
              if applicable)
            * 'taper' - Taper source spectrum

    Raises
    ------
    synphot.exceptions.DisjointError
        Bandpass does not overlap with source spectrum.

    synphot.exceptions.PartialOverlap
        Bandpass only partially overlaps with source spectrum
        when they must fully overlap.

    synphot.exceptions.SynphotError
        Invalid inputs.

    synphot.exceptions.UndefinedBinset
        Missing binned wavelength set.

    """
    def __init__(self, spec, band, binset=None, force='none'):
        # Duck-type specutils.Spectrum1D to avoid hard dependency on specutils
        if hasattr(spec, 'flux') and hasattr(spec, 'spectral_axis'):
            spec = SourceSpectrum.from_spectrum1d(spec)

        if not isinstance(spec, SourceSpectrum):
            raise exceptions.SynphotError('Invalid source spectrum.')

        if not isinstance(band, SpectralElement):
            raise exceptions.SynphotError('Invalid bandpass.')

        # Inherit input warnings like ASTROLIB PYSYNPHOT
        warn = {}

        # Validate overlap
        if force is None:
            force = 'none'
        else:
            force = force.lower()
        stat = band.check_overlap(spec)

        if stat == 'none':
            raise exceptions.DisjointError(
                'Source spectrum and bandpass are disjoint.')

        elif 'partial' in stat:

            if force == 'none':
                raise exceptions.PartialOverlap(
                    'Source spectrum and bandpass do not fully overlap. '
                    'You may use force=[extrap|taper] to force this '
                    'Observation anyway.')

            elif force == 'taper':
                spec = spec.taper()
                msg = 'Source spectrum is tapered.'
                warnings.warn(msg, AstropyUserWarning)
                warn['PartialOverlap'] = msg

            elif force.startswith('extrap'):
                if spec.force_extrapolation():
                    msg = ('Source spectrum will be extrapolated (at constant '
                           'value for empirical model).')
                else:
                    msg = ('Source spectrum will be evaluated outside '
                           'pre-defined waveset.')

                warnings.warn(msg, AstropyUserWarning)
                warn['PartialOverlap'] = msg

            else:
                raise exceptions.SynphotError(
                    'force={0} is invalid, must be "none", "taper", '
                    'or "extrap"'.format(force))

        elif stat != 'full':  # pragma: no cover
            raise exceptions.SynphotError(
                'Overlap result of {0} is unexpected'.format(stat))

        # Create composite spectrum
        super(Observation, self).__init__(spec * band, clean_meta=True)
        self._spec = spec
        self._band = band
        self._force = force

        # Merge in other warnings
        self.warnings = warn

        # Initialize bins
        self._init_bins(binset)

    def _init_bins(self, binset):
        """Calculated binned wavelength centers, edges, and flux.

        By contrast, the native waveset and flux should be considered
        samples of a continuous function.

        Thus, it makes sense to interpolate ``self.waveset`` and
        ``self(self.waveset)``, but not `binset` and `binflux`.

        """
        if binset is None:
            if self.bandpass.waveset is not None:
                self._binset = self.bandpass.waveset
            elif self.spectrum.waveset is not None:
                self._binset = self.spectrum.waveset
                log.info('Bandpass waveset is undefined; '
                         'Using source spectrum waveset instead.')
            else:
                raise exceptions.UndefinedBinset(
                    'Both source spectrum and bandpass have undefined '
                    'waveset; Provide binset manually.')
        else:
            self._binset = self._validate_wavelengths(binset)

        # binset must be in ascending order for calcbinflux()
        # to work properly.
        if self._binset[0] > self._binset[-1]:
            self._binset = self._binset[::-1]

        self._bin_edges = binning.calculate_bin_edges(self._binset)

        # Merge bin edges and centers in with the natural waveset
        spwave = utils.merge_wavelengths(
            self._bin_edges.value, self._binset.value)
        if self.waveset is not None:
            spwave = utils.merge_wavelengths(spwave, self.waveset.value)

        # Throw out invalid wavelengths after merging.
        spwave = spwave[spwave > 0]

        # Compute indices associated to each endpoint.
        indices = np.searchsorted(spwave, self._bin_edges.value)
        i_beg = indices[:-1]
        i_end = indices[1:]

        # Prepare integration variables.
        flux = self(spwave)
        avflux = (flux.value[1:] + flux.value[:-1]) * 0.5
        deltaw = spwave[1:] - spwave[:-1]

        # Sum over each bin.
        binflux, intwave = binning.calcbinflux(
            self._binset.size, i_beg, i_end, avflux, deltaw)

        self._binflux = binflux * flux.unit

    @property
    def spectrum(self):
        """Source spectrum of the observation."""
        return self._spec

    @property
    def bandpass(self):
        """Bandpass of the observation."""
        return self._band

    @property
    def binset(self):
        """Center of binned wavelengths."""
        return self._binset

    @property
    def bin_edges(self):
        """Edges of binned wavelengths."""
        return self._bin_edges

    @property
    def binflux(self):
        """Binned flux corresponding to `binset`."""
        return self._binflux

    def __mul__(self, other):
        """Multiply self and other."""
        sp = self.spectrum * other
        obs = self.__class__(
            sp, self.bandpass, binset=self.binset, force=self._force)
        return obs

    def taper(self, **kwargs):
        """Tapering is disabled."""
        raise NotImplementedError('Observation cannot be tapered.')

    def _validate_binned_wavelengths(self, wave):
        if wave is None:
            wavelengths = self.binset
        else:
            wavelengths = self._validate_wavelengths(wave)
        return wavelengths

    def sample_binned(self, wavelengths=None, flux_unit=None, **kwargs):
        """Sample binned observation without interpolation.

        To sample unbinned data, use ``__call__``.

        Parameters
        ----------
        wavelengths : array-like, `~astropy.units.quantity.Quantity`, or `None`
            Wavelength values for sampling.
            If not a Quantity, assumed to be in Angstrom.
            If `None`, `binset` is used.

        flux_unit : str or `~astropy.units.Unit` or `None`
            Flux is converted to this unit.
            If not given, internal unit is used.

        kwargs : dict
            Keywords acceptable by :func:`~synphot.units.convert_flux`.

        Returns
        -------
        flux : `~astropy.units.quantity.Quantity`
            Binned flux in given unit.

        Raises
        ------
        synphot.exceptions.InterpolationNotAllowed
            Interpolation of binned data is not allowed.

        """
        x = self._validate_binned_wavelengths(wavelengths)
        i = np.searchsorted(self.binset, x)
        if not np.allclose(self.binset[i].value, x.value):
            raise exceptions.InterpolationNotAllowed(
                'Some or all wavelength values are not in binset.')
        y = self.binflux[i]

        if flux_unit is None:
            flux = y
        else:
            flux = units.convert_flux(x, y, flux_unit, **kwargs)

        return flux

    def _get_binned_arrays(self, wavelengths, flux_unit, area=None,
                           vegaspec=None):
        """Get binned observation in user units."""
        x = self._validate_binned_wavelengths(wavelengths)
        y = self.sample_binned(wavelengths=x, flux_unit=flux_unit, area=area,
                               vegaspec=vegaspec)

        if isinstance(wavelengths, u.Quantity):
            w = x.to(wavelengths.unit, u.spectral())
        else:
            w = x

        return w, y

    def binned_waverange(self, cenwave, npix, **kwargs):
        """Calculate the wavelength range covered by the given number
        of pixels centered on the given central wavelengths of
        `binset`.

        Parameters
        ----------
        cenwave : float or `~astropy.units.quantity.Quantity`
            Desired central wavelength.
            If not a Quantity, assumed to be in Angstrom.

        npix : int
            Desired number of pixels, centered on ``cenwave``.

        kwargs : dict
            Keywords accepted by :func:`synphot.binning.wave_range`.

        Returns
        -------
        waverange : `~astropy.units.quantity.Quantity`
            Lower and upper limits of the wavelength range,
            in the unit of ``cenwave``.

        """
        # Calculation is done in the unit of cenwave.
        if not isinstance(cenwave, u.Quantity):
            cenwave = cenwave * self._internal_wave_unit

        bin_wave = units.validate_quantity(
            self.binset, cenwave.unit, equivalencies=u.spectral())

        return binning.wave_range(
            bin_wave.value, cenwave.value, npix, **kwargs) * cenwave.unit

    def binned_pixelrange(self, waverange, **kwargs):
        """Calculate the number of pixels within the given wavelength
        range and `binset`.

        Parameters
        ----------
        waverange : tuple of float or `~astropy.units.quantity.Quantity`
            Lower and upper limits of the desired wavelength range.
            If not a Quantity, assumed to be in Angstrom.

        kwargs : dict
            Keywords accepted by :func:`synphot.binning.pixel_range`.

        Returns
        -------
        npix : int
            Number of pixels.

        """
        x = units.validate_quantity(
            waverange, self._internal_wave_unit, equivalencies=u.spectral())
        return binning.pixel_range(self.binset.value, x.value, **kwargs)

    def effective_wavelength(self, binned=True, wavelengths=None,
                             mode='efflerg'):
        """Calculate :ref:`effective wavelength <synphot-formula-effwave>`.

        Parameters
        ----------
        binned : bool
            Sample data in native wavelengths if `False`.
            Else, sample binned data (default).

        wavelengths : array-like, `~astropy.units.quantity.Quantity`, or `None`
            Wavelength values for sampling.
            If not a Quantity, assumed to be in Angstrom.
            If `None`, ``self.waveset`` or `binset` is used, depending
            on ``binned``.

        mode : {'efflerg', 'efflphot'}
            Flux is first converted to the unit below before calculation:

                * 'efflerg' - FLAM
                * 'efflphot' - PHOTLAM (deprecated)

        Returns
        -------
        eff_lam : `~astropy.units.quantity.Quantity`
            Observation effective wavelength.

        Raises
        ------
        synphot.exceptions.SynphotError
            Invalid mode.

        """
        mode = mode.lower()
        if mode == 'efflerg':
            flux_unit = units.FLAM
        elif mode == 'efflphot':
            warnings.warn(
                'Usage of EFFLPHOT is deprecated.', AstropyDeprecationWarning)
            flux_unit = units.PHOTLAM
        else:
            raise exceptions.SynphotError(
                'mode must be "efflerg" or "efflphot"')

        if binned:
            x = self._validate_binned_wavelengths(wavelengths).value
            y = self.sample_binned(wavelengths=x, flux_unit=flux_unit).value
        else:
            x = self._validate_wavelengths(wavelengths).value
            y = units.convert_flux(x, self(x), flux_unit).value

        num = trapezoid(y * x ** 2, x=x)
        den = trapezoid(y * x, x=x)

        if den == 0.0:  # pragma: no cover
            eff_lam = 0.0
        else:
            eff_lam = abs(num / den)

        return eff_lam * self._internal_wave_unit

    # https://github.com/spacetelescope/synphot_refactor/issues/159
    def effstim(self, flux_unit=None, wavelengths=None, area=None,
                vegaspec=None):
        """Calculate :ref:`effective stimulus <synphot-formula-effstim>`
        for given flux unit.

        Parameters
        ----------
        flux_unit : str or `~astropy.units.Unit` or `None`
            The unit of effective stimulus.
            COUNT gives result in count/s (see :meth:`countrate` for more
            options).
            If not given, internal unit is used.

        wavelengths : array-like, `~astropy.units.quantity.Quantity`, or `None`
            Wavelength values for sampling. This must be given if
            ``self.waveset`` is undefined for the underlying spectrum model(s).
            If not a Quantity, assumed to be in Angstrom.
            If `None`, ``self.waveset`` is used.

        area, vegaspec
            See :func:`~synphot.units.convert_flux`.

        Returns
        -------
        eff_stim : `~astropy.units.quantity.Quantity`
            Observation effective stimulus based on given flux unit.

        """
        if flux_unit is None:
            flux_unit = self._internal_flux_unit

        flux_unit = units.validate_unit(flux_unit)
        flux_unit_name = flux_unit.to_string()

        # Special handling of COUNT/OBMAG.
        # This is special case of countrate calculations.
        if flux_unit == u.count or flux_unit_name == units.OBMAG.to_string():
            val = self.countrate(area, binned=False, wavelengths=wavelengths)

            if flux_unit == units.OBMAG:
                eff_stim = (-2.5 * np.log10(val.value)) * flux_unit
            else:
                eff_stim = val

            return eff_stim

        # Special handling of VEGAMAG.
        # This is basically effstim(self)/effstim(Vega)
        if flux_unit_name == units.VEGAMAG.to_string():
            if not isinstance(vegaspec, SourceSpectrum):
                raise exceptions.SynphotError('Vega spectrum is missing.')
            num = self.integrate(wavelengths=wavelengths)
            den = (vegaspec * self.bandpass).integrate(
                integration_type='trapezoid')
            utils.validate_totalflux(num)
            utils.validate_totalflux(den)
            return (2.5 * (math.log10(den.value) -
                           math.log10(num.value))) * units.VEGAMAG

        # Sample the bandpass
        x_band = self.bandpass._validate_wavelengths(wavelengths).value
        y_band = self.bandpass(x_band).value

        # Sample the observation in FLAM
        inwave = self._validate_wavelengths(wavelengths).value
        influx = units.convert_flux(inwave, self(inwave), units.FLAM).value

        # Integrate
        num = abs(trapezoid(inwave * influx, x=inwave))
        den = abs(trapezoid(x_band * y_band, x=x_band))
        utils.validate_totalflux(num)
        utils.validate_totalflux(den)
        val = (num / den) * units.FLAM

        # Integration should always be done in FLAM and then
        # converted to desired units as follows.
        if flux_unit.physical_type == 'spectral flux density wav':
            if flux_unit == u.STmag:
                eff_stim = val.to(flux_unit)
            else:  # FLAM
                eff_stim = val
        elif flux_unit.physical_type in (
                'spectral flux density', 'photon flux density',
                'photon flux density wav'):
            w_pivot = self.bandpass.pivot()
            eff_stim = units.convert_flux(w_pivot, val, flux_unit)
        else:
            raise exceptions.SynphotError(
                'Flux unit {0} is invalid'.format(flux_unit))

        return eff_stim

    def countrate(self, area, binned=True, wavelengths=None, waverange=None,
                  force=False):
        """Calculate :ref:`effective stimulus <synphot-formula-effstim>`
        in count/s.

        Parameters
        ----------
        area : float or `~astropy.units.quantity.Quantity`
            Area that flux covers. If not a Quantity, assumed to be in
            :math:`cm^{2}`.

        binned : bool
            Sample data in native wavelengths if `False`.
            Else, sample binned data (default).

        wavelengths : array-like, `~astropy.units.quantity.Quantity`, or `None`
            Wavelength values for sampling. This must be given if
            ``self.waveset`` is undefined for the underlying spectrum model(s).
            If not a Quantity, assumed to be in Angstrom.
            If `None`, ``self.waveset`` or `binset` is used, depending
            on ``binned``.

        waverange : tuple of float, Quantity, or `None`
            Lower and upper limits of the desired wavelength range.
            If not a Quantity, assumed to be in Angstrom.
            If `None`, the full range is used.

        force : bool
            If a wavelength range is given, partial overlap raises
            an exception when this is `False` (default). Otherwise,
            it returns calculation for the overlapping region.
            Disjoint wavelength range raises an exception regardless.

        Returns
        -------
        count_rate : `~astropy.units.quantity.Quantity`
            Observation effective stimulus in count/s.

        Raises
        ------
        synphot.exceptions.DisjointError
            Wavelength range does not overlap with observation.

        synphot.exceptions.PartialOverlap
            Wavelength range only partially overlaps with observation.

        synphot.exceptions.SynphotError
            Calculation failed, including but not limited to NaNs in flux.

        """
        # Sample the observation
        if binned:
            x = self._validate_binned_wavelengths(wavelengths).value
            y = self.sample_binned(wavelengths=x, flux_unit=u.count,
                                   area=area).value
        else:
            x = self._validate_wavelengths(wavelengths).value
            y = units.convert_flux(x, self(x), u.count,
                                   area=area).value

        # Use entire wavelength range
        if waverange is None:
            influx = y

        # Use given wavelength range
        else:
            w = units.validate_quantity(waverange, self._internal_wave_unit,
                                        equivalencies=u.spectral()).value
            stat = utils.overlap_status(w, x)
            w1 = w.min()
            w2 = w.max()

            if stat == 'none':
                raise exceptions.DisjointError(
                    'Observation and wavelength range are disjoint.')
            elif 'partial' in stat:
                if force:
                    warnings.warn(
                        'Count rate calculated only for wavelengths in the '
                        'overlap between observation and given range.',
                        AstropyUserWarning)
                    w1 = max(w1, x.min())
                    w2 = min(w2, x.max())
                else:
                    raise exceptions.PartialOverlap(
                        'Observation and wavelength range do not fully '
                        'overlap. You may use force=True to force this '
                        'calculation anyway.')
            elif stat != 'full':  # pragma: no cover
                raise exceptions.SynphotError(
                    'Overlap result of {0} is unexpected'.format(stat))

            if binned:
                if wavelengths is None:
                    bin_edges = self.bin_edges.value
                else:
                    bin_edges = binning.calculate_bin_edges(x).value
                i1 = np.searchsorted(bin_edges, w1) - 1
                i2 = np.searchsorted(bin_edges, w2)
                influx = y[i1:i2]
            else:
                mask = ((x >= w1) & (x <= w2))
                influx = y[mask]

        val = math.fsum(influx)
        utils.validate_totalflux(val)

        return val * (u.count / u.s)

    def plot(self, binned=True, wavelengths=None, flux_unit=None, area=None,
             vegaspec=None, **kwargs):  # pragma: no cover
        """Plot the observation.

        .. note:: Uses ``matplotlib``.

        Parameters
        ----------
        binned : bool
            Plot data in native wavelengths if `False`.
            Else, plot binned data (default).

        wavelengths : array-like, `~astropy.units.quantity.Quantity`, or `None`
            Wavelength values for sampling.
            If not a Quantity, assumed to be in Angstrom.
            If `None`, ``self.waveset`` or `binset` is used, depending
            on ``binned``.

        flux_unit : str or `~astropy.units.Unit` or `None`
            Flux is converted to this unit for plotting.
            If not given, internal unit is used.

        area, vegaspec
            See :func:`~synphot.units.convert_flux`.

        kwargs : dict
            See :func:`synphot.spectrum.BaseSpectrum.plot`.

        Raises
        ------
        synphot.exceptions.SynphotError
            Invalid inputs.

        """
        if binned:
            w, y = self._get_binned_arrays(wavelengths, flux_unit, area=area,
                                           vegaspec=vegaspec)
        else:
            w, y = self._get_arrays(wavelengths, flux_unit=flux_unit,
                                    area=area, vegaspec=vegaspec)
        self._do_plot(w, y, **kwargs)

    def as_spectrum(self, binned=True, wavelengths=None):
        """Reduce the observation to an empirical source spectrum.

        An observation is a complex object with some restrictions
        on its capabilities. At times, it would be useful to work
        with the observation as a simple object that is easier to
        manipulate and takes up less memory.

        This is also useful for writing an observation as sampled
        spectrum out to a FITS file.

        Parameters
        ----------
        binned : bool
            Write out data in native wavelengths if `False`.
            Else, write binned data (default).

        wavelengths : array-like, `~astropy.units.quantity.Quantity`, or `None`
            Wavelength values for sampling.
            If not a Quantity, assumed to be in Angstrom.
            If `None`, ``self.waveset`` or `binset` is used, depending
            on ``binned``.

        Returns
        -------
        sp : `~synphot.spectrum.SourceSpectrum`
            Empirical source spectrum.

        """
        if binned:
            w, y = self._get_binned_arrays(
                wavelengths, self._internal_flux_unit)
        else:
            w, y = self._get_arrays(
                wavelengths, flux_unit=self._internal_flux_unit)

        header = {'observation': str(self), 'binned': binned}
        return SourceSpectrum(Empirical1D, points=w, lookup_table=y,
                              meta={'header': header})



# Licensed under a 3-clause BSD style license - see LICENSE.rst
"""Test thermal.py module."""

# STDLIB
import os

# THIRD-PARTY
import numpy as np
import pytest

# ASTROPY
from astropy import units as u
from astropy.utils.data import get_pkg_data_filename

# LOCAL
from synphot import exceptions
from synphot.thermal import ThermalSpectralElement


def setup_module(module):
    import astropy.constants as const
    from astropy.constants import si, astropyconst13

    const.sigma_sb = si.sigma_sb = astropyconst13.sigma_sb
    const.h = si.h = astropyconst13.h
    const.k_B = si.k_B = astropyconst13.k_B


def teardown_module(module):
    import astropy.constants as const
    from astropy.constants import si, astropyconst20

    const.sigma_sb = si.sigma_sb = astropyconst20.sigma_sb
    const.h = si.h = astropyconst20.h
    const.k_B = si.k_B = astropyconst20.k_B


class TestThermalSpectralElement:
    """Test ``ThermalSpectralElement``."""
    def setup_class(self):
        thfile = get_pkg_data_filename(
            os.path.join('data', 'wfc3_ir_g141_src_003_th.fits'),
            package='synphot.tests')
        self.th = ThermalSpectralElement.from_file(thfile)

    def test_taper(self):
        with pytest.raises(NotImplementedError):
            self.th.taper()

    def test_properties(self):
        assert self.th.temperature == 237.3 * u.K
        assert self.th.beam_fill_factor == 1

    def test_thermal_source(self):
        sp = self.th.thermal_source()
        assert sp.meta['temperature'] == self.th.temperature
        assert sp.meta['beam_fill_factor'] == self.th.beam_fill_factor
        np.testing.assert_allclose(
            sp([6800, 7800, 8800, 17920, 18920, 19920]).value,
            [1.246735e-30, 6.63655885e-26, 2.80933935e-22, 2.76427032e-08,
             1.33011769e-07, 5.40857951e-07], rtol=1e-5)

    def test_from_file_exceptions(self):
        # Non-FITS file
        with pytest.raises(exceptions.SynphotError):
            ThermalSpectralElement.from_file('dummy.txt')

        # Missing DEFT keyword
        thfile = get_pkg_data_filename(
            os.path.join('data', 'hst_acs_hrc_f555w.fits'),
            package='synphot.tests')
        with pytest.raises(exceptions.SynphotError):
            ThermalSpectralElement.from_file(thfile, flux_col='THROUGHPUT')



#!/usr/bin/env python
# Licensed under a 3-clause BSD style license - see LICENSE.rst

import os
import sys
from setuptools import setup, Extension


def get_extensions():
    from collections import defaultdict
    import numpy

    cfg = defaultdict(list)
    cfg['include_dirs'].extend([
        numpy.get_include(),
        os.path.join('synphot', 'include')])
    cfg['sources'] = [
        os.path.join('synphot', 'src', 'synphot_utils.c')]
    cfg = dict((str(key), val) for key, val in cfg.items())

    return [Extension('synphot.synphot_utils', **cfg)]


TEST_HELP = """
Note: running tests is no longer done using 'python setup.py test'. Instead
you will need to run:

    pip install -e .
    pytest

For more information, see:

  https://docs.astropy.org/en/latest/development/testguide.html#running-tests
"""

if 'test' in sys.argv:
    print(TEST_HELP)
    sys.exit(1)

DOCS_HELP = """
Note: building the documentation is no longer done using
'python setup.py build_docs'. Instead you will need to run:

    cd docs
    make html

For more information, see:

  https://docs.astropy.org/en/latest/install.html#builddocs
"""

if 'build_docs' in sys.argv or 'build_sphinx' in sys.argv:
    print(DOCS_HELP)
    sys.exit(1)

# Note that requires and provides should not be included in the call to
# ``setup``, since these are now deprecated. See this link for more details:
# https://groups.google.com/forum/#!topic/astropy-dev/urYO8ckB2uM

setup(use_scm_version={'write_to': 'synphot/version.py'},
      ext_modules=get_extensions())



# Licensed under a 3-clause BSD style license - see LICENSE.rst
"""Test different integration methods.
Things not covered here but somewhere else:

* ``conf.set_temp('default_integrator', 'analytical')`` and its default
  covered in ``spectrum.rst``.
* The following integrations are tested in ``test_spectrum.py``:
    * ``BlackBodyNorm1D`` source, and indirectly, ``BlackBody1D``.
    * ``Box1D`` bandpass.
    * ``Empirical1D`` source and bandpass.
    * ``GaussianFlux1D`` source.
    * ``PowerLawFlux1D`` source.
    * Compound source and bandpass.

"""

# THIRD-PARTY
import numpy as np
import pytest
from astropy import units as u
from astropy.modeling.models import Const1D
from astropy.tests.helper import assert_quantity_allclose
from numpy.testing import assert_allclose
from scipy.integrate import trapezoid

# LOCAL
from synphot import models, units
from synphot.exceptions import SynphotError
from synphot.spectrum import SourceSpectrum, SpectralElement


class TestSourceConstFlux1D:
    def setup_class(self):
        self.sp = SourceSpectrum(models.ConstFlux1D, amplitude=1 * units.FLAM)
        self.w = np.arange(5499, 5600) * u.AA
        self.ans_flam = 100 * (u.erg / (u.cm * u.cm * u.s))
        self.ans_photlam = 2.79343128e+13 * (u.ph / (u.cm * u.cm * u.s))

    def test_integrate(self):
        assert_quantity_allclose(
            self.sp.integrate(wavelengths=self.w), self.ans_photlam)
        assert_quantity_allclose(
            self.sp.integrate(wavelengths=self.w, flux_unit='flam'),
            self.ans_flam)
        assert_quantity_allclose(
            self.sp.integrate(wavelengths=self.w, flux_unit='flam',
                              integration_type='analytical'),
            self.ans_flam)

        with pytest.raises(SynphotError, match='cannot operate in Jy'):
            self.sp.integrate(wavelengths=self.w, flux_unit=u.Jy)

    @pytest.mark.xfail(reason='Cannot convert to PHOTLAM')
    def test_integrate_wontfix(self):
        """Constant spectrum in FLAM is not constant in PHOTLAM, so cannot
        apply simple unit conversion to integrated flux.
        """
        assert_quantity_allclose(
            self.sp.integrate(wavelengths=self.w,
                              integration_type='analytical'),
            self.ans_photlam)

    def test_integrate_freq_space(self):
        """When flux is constant in frequency space."""
        sp = SourceSpectrum(models.ConstFlux1D, amplitude=1 * u.Jy)
        nu = self.w.to(u.Hz, u.spectral())

        # Answer in this setup is basically
        # ((max(nu) - min(nu)) * u.Jy).to(u.erg / (u.cm * u.cm * u.s))
        ans_flam = 9.73703017e-11 * (u.erg / (u.cm * u.cm * u.s))

        assert_quantity_allclose(
            sp.integrate(wavelengths=self.w, flux_unit='flam'), ans_flam)
        assert_quantity_allclose(
            sp.integrate(wavelengths=self.w, flux_unit='flam',
                         integration_type='analytical'),
            ans_flam)
        assert_quantity_allclose(
            sp.integrate(wavelengths=nu, flux_unit='flam'), ans_flam)
        assert_quantity_allclose(
            sp.integrate(wavelengths=nu, flux_unit='flam',
                         integration_type='analytical'),
            ans_flam)


def test_bandpass_Const1D():
    """Test Const1D bandpass and a few optional logic routes."""
    bp = SpectralElement(Const1D, amplitude=1)
    w = np.arange(5499, 5600) * u.AA
    ans = 100 * u.AA

    assert_quantity_allclose(bp.integrate(wavelengths=w), ans)

    # Const1D from astropy has no integrate method, so this silently falls
    # back to trapezoid integration anyway.
    assert_quantity_allclose(
        bp.integrate(wavelengths=w, integration_type='analytical'), ans)

    with pytest.raises(SynphotError, match='waveset is undefined'):
        bp.integrate()

    with pytest.raises(SynphotError, match='flux_unit cannot be used'):
        bp.integrate(wavelengths=w, flux_unit='flam')

    with pytest.raises(NotImplementedError,
                       match='not a supported integration type'):
        bp.integrate(wavelengths=w, integration_type='notreal')


def test_bandpass_Gaussian1D():
    bp = SpectralElement(
        models.Gaussian1D, amplitude=0.8, mean=5000, stddev=130)
    ans = 260.68913672 * u.AA
    assert_quantity_allclose(bp.integrate(), ans)
    assert_quantity_allclose(bp.integrate(integration_type='analytical'), ans,
                             rtol=1e-6)

    # Equivalent width is basically the integral.
    assert_quantity_allclose(bp.equivwidth(), ans)
    assert_quantity_allclose(bp.equivwidth(integration_type='analytical'), ans,
                             rtol=1e-6)


def test_bandpass_Lorentz1D():
    bp = SpectralElement(models.Lorentz1D, amplitude=0.8, x_0=5000, fwhm=130)
    ans = 161.28101053 * u.AA
    assert_quantity_allclose(bp.integrate(), ans)
    assert_quantity_allclose(bp.integrate(integration_type='analytical'), ans)


def test_source_Lorentz1D():
    sp = SourceSpectrum(
        models.Lorentz1D, amplitude=1 * units.PHOTLAM, x_0=5000, fwhm=130)
    ans_photlam = 201.60126316 * (u.ph / (u.cm * u.cm * u.s))
    ans_flam = 8.06013348e-10 * (u.erg / (u.cm * u.cm * u.s))

    assert_quantity_allclose(sp.integrate(), ans_photlam)
    assert_quantity_allclose(sp.integrate(flux_unit='flam'), ans_flam)
    assert_quantity_allclose(
        sp.integrate(integration_type='analytical'), ans_photlam)

    # Flux unit conversion is iffy here because we do not have proper
    # LorentzFlux1D model.
    assert_quantity_allclose(
        sp.integrate(flux_unit='flam', integration_type='analytical'),
        ans_flam, rtol=0.01)


def test_bandpass_RickerWavelet1D():
    """Does not make much sense when transmission is negative, so this
    makes more sense as equivalent width calculation, which is also the
    integral.
    """
    bp = SpectralElement(
        models.RickerWavelet1D, amplitude=1, x_0=5000, sigma=100)
    ans = 242.20770777 * u.AA
    assert_quantity_allclose(bp.equivwidth(), ans)
    assert_quantity_allclose(bp.equivwidth(integration_type='analytical'),
                             ans, rtol=5e-3)

    with pytest.raises(NotImplementedError, match='Partial analytic'):
        bp.equivwidth(integration_type='analytical',
                      wavelengths=np.arange(4900, 5200) * u.AA)

    with pytest.raises(NotImplementedError, match='Partial analytic'):
        bp.equivwidth(integration_type='analytical',
                      wavelengths=np.arange(4800, 5100) * u.AA)


def test_source_RickerWavelet1D():
    pytest.xfail(
        'Not sure if this test makes sense as negative flux is unphysical')


def test_bandpass_Trapezoid1D():
    """Trapezoid integration of a trapezoid. How meta!"""
    bp = SpectralElement(
        models.Trapezoid1D, amplitude=0.8, x_0=5000, width=10, slope=0.25)
    ans = 10.56 * u.AA
    assert_quantity_allclose(bp.integrate(), ans)
    assert_quantity_allclose(bp.integrate(integration_type='analytical'), ans)


def test_trapezoid_box1d():
    """Test the underlying trapezoid integration."""
    m = models.Box1D(amplitude=1, x_0=5000, width=10)

    # Ascending.
    x = m.sampleset()
    assert_allclose(trapezoid(m(x), x=x), 10)

    # Descending.
    x2 = x[::-1]
    assert_allclose(abs(trapezoid(m(x2), x=x2)), 10)



# Licensed under a 3-clause BSD style license - see LICENSE.rst
"""Test utils.py module."""

# STDLIB
import os

# THIRD PARTY
import numpy as np
import pytest

# ASTROPY
from astropy import units as u

# LOCAL
from synphot import exceptions, utils, units


@pytest.mark.parametrize(
    ('a', 'b', 'ans'),
    [(np.arange(5, 8), np.arange(10), 'full'),
     (np.arange(10), np.arange(5, 8), 'partial'),
     (np.arange(3), np.arange(2, 5), 'partial'),
     (np.arange(4, 8), np.arange(2, 5), 'partial'),
     (np.arange(3), np.arange(3, 6), 'none'),
     (np.arange(3, 6), np.arange(3), 'none')])
def test_overlap_status(a, b, ans):
    """Test overlap status validation."""
    assert utils.overlap_status(a, b) == ans


def test_validate_totalflux():
    """Test integrated flux validation."""
    utils.validate_totalflux(0.01)
    utils.validate_totalflux(0.01 * units.FLAM)


@pytest.mark.filterwarnings('ignore:invalid value')
@pytest.mark.parametrize(
    'val', (-0.01, -0.01 * units.FLAM,
            0, 0 * units.PHOTLAM,
            np.inf, np.inf * u.Jy,
            np.nan, np.nan * units.FLAM))
def test_validate_totalflux_invalid(val):
    """Invalid integrated flux."""
    with pytest.raises(exceptions.SynphotError):
        utils.validate_totalflux(val)


def test_validate_wavelengths():
    """Test wavelengths validation."""
    # Valid wavelengths (ascending)
    a = np.arange(1, 11)
    utils.validate_wavelengths(a)

    # Valid wavelengths (descending)
    a = a[::-1]
    utils.validate_wavelengths(a * u.micron)

    # Invalid wavelengths
    with pytest.raises(exceptions.SynphotError):
        utils.validate_wavelengths(1.0 * u.K)
    with pytest.raises(exceptions.ZeroWavelength):
        utils.validate_wavelengths(np.arange(10))
    with pytest.raises(exceptions.UnsortedWavelength):
        utils.validate_wavelengths([1000, 1002, 1001, 1003, 1004])

    try:
        utils.validate_wavelengths([1000, 1001, 1002, 1003, 1003])
    except exceptions.DuplicateWavelength as e:
        np.testing.assert_array_equal(e.rows, 3)


@pytest.mark.parametrize(
    ('num', 'delta', 'log', 'ans'),
    [(10, None, True, [10.0, 10.71773463, 11.48698355, 12.31144413,
                       13.19507911, 14.14213562, 15.15716567, 16.24504793,
                       17.41101127, 18.66065983]),
     (0, 0.05, True, [10.0, 11.22018454, 12.58925412, 14.12537545, 15.84893192,
                      17.7827941, 19.95262315]),
     (10, None, False, np.arange(10, 20)),
     (0, 1.0, False, np.arange(10, 20))])
def test_genwave(num, delta, log, ans):
    """Test wavelength generation."""
    wave, wave_str = utils.generate_wavelengths(
        minwave=10, maxwave=20, num=num, delta=delta, log=log,
        wave_unit=u.micron)
    np.testing.assert_allclose(wave.value, ans)
    assert wave.unit == u.micron
    assert isinstance(wave_str, str)


class TestMergeWave:
    """Test wavelengths merging."""
    def setup_class(self):
        self.thres = 1e-12
        self.wave = [5000.0, 5000.01, 5000.02, 5000.03, 5000.04, 6000.0]

    def test_merge_none(self):
        assert utils.merge_wavelengths(None, None) is None
        np.testing.assert_array_equal(
            utils.merge_wavelengths(None, self.wave), self.wave)
        np.testing.assert_array_equal(
            utils.merge_wavelengths(self.wave, None), self.wave)

    def test_merge_thres(self):
        w = [5000.005, 5000.02 + self.thres, 5500.0, 6000.0]
        ans = [5000.0, 5000.005, 5000.01, 5000.02, 5000.03, 5000.04, 5500.0,
               6000.0]
        wave = utils.merge_wavelengths(self.wave, w, threshold=self.thres)
        dw = wave[1:] - wave[:-1]
        np.testing.assert_allclose(wave, ans)
        assert np.all(dw > self.thres)

    def test_merge_same(self):
        wave = utils.merge_wavelengths(self.wave, self.wave)
        np.testing.assert_array_equal(wave, self.wave)


def test_download_bad_root(tmp_path):
    """Test data download helper when input dir is invalid."""
    ptr = tmp_path / 'bad_cdbs'
    ptr.write_text("something")
    cdbs_root = str(ptr)

    with pytest.raises(OSError, match=".* must be a directory"):
        utils.download_data(cdbs_root, verbose=False)

    with pytest.raises(FileNotFoundError):
        utils.download_data('', verbose=False)


def test_download_data(tmp_path):
    """Test data download helper in dry run mode."""
    from synphot.config import conf

    # Use case where user downloads all data into new dir.
    cdbs_root = str(tmp_path / 'cdbs')
    file_list_1 = utils.download_data(cdbs_root, verbose=False, dry_run=True)
    filename = file_list_1[0]
    assert len(file_list_1) == 21
    assert filename.startswith(cdbs_root)
    assert os.path.isdir(os.path.join(cdbs_root, 'calspec'))

    # Make dummy files for the next step.
    for fname in file_list_1:
        with open(fname, 'w') as f:
            f.write('\n')

    # Use case where user downloads only some data into existing dir.
    os.remove(filename)
    file_list_2 = utils.download_data(cdbs_root, verbose=False, dry_run=True)
    assert len(file_list_2) == 1 and file_list_2[0] == filename

    # Re-create the deleted dummy file for next step.
    with open(filename, 'w') as f:
        f.write('\n')

    # Use case where user redefined data file to be non-STScI.
    # While the given file will be used, default Vega is downloaded anyway.
    filename = [fname for fname in file_list_1
                if fname.endswith('alpha_lyr_stis_010.fits')][0]
    os.remove(filename)
    with conf.set_temp('vega_file', '/custom/host/my_vega.fits'):
        file_list_2 = utils.download_data(
            cdbs_root, verbose=False, dry_run=True)
    assert len(file_list_2) == 1 and file_list_2[0] == filename


def test_download_data_cache_only():
    """Test data download helper in dry run mode (cache only)."""

    # Use case where user downloads all data into new dir.
    # The other use cases in test_download_data() will depend on the native
    # behavior of astropy caching and not tested here.
    file_list_1 = utils.download_data(None, verbose=False, dry_run=True)
    filename = file_list_1[0]
    assert len(file_list_1) == 21
    assert filename.startswith('http')



# Licensed under a 3-clause BSD style license - see LICENSE.rst
"""Test spectrum.py module and related functionalities that are not covered
by ``test_spectrum_source.py`` nor ``test_spectrum_bandpass.py``."""

# THIRD-PARTY
import numpy as np
import pytest

# ASTROPY
from astropy import units as u
from astropy.io import fits
from astropy.modeling.models import Const1D, RedshiftScaleFactor
from astropy.tests.helper import assert_quantity_allclose

# LOCAL
from synphot.tests.test_units import _wave, _flux_jy, _flux_photlam
from synphot import exceptions, units
from synphot.models import Box1D, Empirical1D, GaussianFlux1D, get_waveset
from synphot.spectrum import SourceSpectrum, SpectralElement


def setup_module(module):
    import astropy.constants as const
    from astropy.constants import si, astropyconst13

    const.h = si.h = astropyconst13.h


def teardown_module(module):
    import astropy.constants as const
    from astropy.constants import si, astropyconst40

    const.h = si.h = astropyconst40.h


class TestCheckOverlap:
    """Test spectrum overlap check. This method is ever only used
    in the form of ``bp.check_overlap(sp)``, so that is what is
    tested here.
    """
    def setup_class(self):
        self.bp = SpectralElement(
            Empirical1D, points=[2999.9, 3000, 6000, 6000.1],
            lookup_table=[0, 1, 1, 0])

    def test_full(self):
        """As long as we don't have to extrapolate or taper
        source spectrum, it's okay.
        """
        sp = SourceSpectrum(
            Empirical1D, points=[999.9, 1000, 9000, 9000.1],
            lookup_table=[0, 1, 1, 0])
        assert self.bp.check_overlap(sp) == 'full'

        sp = SourceSpectrum(
            Empirical1D, points=[3999.9, 4000, 4500, 4500.1],
            lookup_table=[0, 1, 1, 0])
        assert self.bp.check_overlap(sp) == 'full'

    def test_partial_most(self):
        """99% overlap."""
        sp = SourceSpectrum(
            Empirical1D, points=[3005, 3005.1, 6000.1, 6000.2],
            lookup_table=[0, 1, 1, 0])
        assert self.bp.check_overlap(sp) == 'partial_most'

    def test_partial_notmost(self):
        """Extrapolation or taper required."""
        sp = SourceSpectrum(
            Empirical1D, points=[3999.9, 4500.1], lookup_table=[1, 1])
        assert self.bp.check_overlap(sp) == 'partial_notmost'

    def test_none(self):
        """No overlap at all."""
        sp = SourceSpectrum(
            Empirical1D, points=[99.9, 100, 2999.8, 2999.9],
            lookup_table=[0, 1, 1, 0])
        assert self.bp.check_overlap(sp) == 'none'

    def test_special_cases(self):
        """One of them has no waveset defined."""
        # Other has no waveset
        sp = SourceSpectrum(Const1D, amplitude=1)
        assert self.bp.check_overlap(sp) == 'full'

        # Self has no waveset
        bp = SpectralElement(Const1D, amplitude=1)
        sp = SourceSpectrum(Box1D, amplitude=1, x_0=5000, width=10)
        assert bp.check_overlap(sp) == 'partial_notmost'

    def test_exceptions(self):
        """Invalid input."""
        with pytest.raises(exceptions.SynphotError):
            self.bp.check_overlap(1)


class TestForceExtrap:
    """Test forcing extrapolation on a source spectrum."""

    @pytest.mark.parametrize('z', [0, 0.03])
    def test_empirical(self, z):
        sp = SourceSpectrum(Empirical1D, points=[1000, 2000, 3000, 4000],
                            lookup_table=[0.5, 0.6, 10.6, 1.5], fill_value=0)
        sp.z = z
        w = [900, 4300]
        assert_quantity_allclose(sp(w), 0 * units.PHOTLAM)  # No extrapolation

        is_forced = sp.force_extrapolation()  # Force extrapolation
        assert is_forced
        assert_quantity_allclose(sp(w), [0.5, 1.5] * units.PHOTLAM)

    def test_analytical(self):
        """Forcing is not possible."""
        sp = SourceSpectrum(GaussianFlux1D, mean=5500, total_flux=1, fwhm=10)
        w = [100, 10000]
        assert_quantity_allclose(sp(w), 0 * units.PHOTLAM)

        is_forced = sp.force_extrapolation()
        assert not is_forced
        assert_quantity_allclose(sp(w), 0 * units.PHOTLAM)


class TestWaveset:
    """Tests related to spectrum waveset."""
    def test_none(self):
        sp = SourceSpectrum(Const1D, amplitude=1)
        assert sp.waveset is None

    def test_sampleset(self):
        tf_unit = u.erg / (u.cm * u.cm * u.s)
        sp = SourceSpectrum(
            GaussianFlux1D, total_flux=(1 * tf_unit), mean=5000, fwhm=10)
        np.testing.assert_array_equal(sp.waveset.value, sp.model.sampleset())

    def test_box1d(self):
        bp = SpectralElement(Box1D, x_0=2000, width=1)
        w = bp.waveset.value
        w_true = bp.model.sampleset()

        np.testing.assert_array_equal(w, w_true)
        np.testing.assert_allclose(
            w[([0, 1, -2, -1], )], bp.model.sampleset(minimal=True))

        # Make sure scale does not change waveset
        bp2 = bp * 2
        bp3 = 0.5 * bp
        np.testing.assert_array_equal(bp2.waveset.value, w_true)
        np.testing.assert_array_equal(bp3.waveset.value, w_true)

    def test_box1d_set_step(self):
        # first test that setting step to default value produces the same
        # waveset as it did before step was a valid input
        bp1 = SpectralElement(Box1D, x_0=2000, width=1)
        assert bp1.waveset.shape == (103,)

        bp2 = SpectralElement(Box1D, x_0=2000, width=1, step=0.01)
        np.testing.assert_array_equal(bp1.waveset.value, bp2.waveset.value)

        # next test that setting a coarser step produces a smaller waveset
        bp3 = SpectralElement(Box1D, x_0=2000, width=1, step=0.2)
        assert bp3.waveset.size < bp1.waveset.size

    def test_composite_none(self):
        bp1 = SpectralElement(Box1D, amplitude=1, x_0=5000, width=10)
        bp2 = SpectralElement(Const1D, amplitude=2)
        bp = bp1 * bp2
        np.testing.assert_array_equal(bp.waveset, bp1.waveset)

    def test_composite(self):
        totflux = 1 * (u.erg / (u.cm * u.cm * u.s))
        g1 = SourceSpectrum(
            GaussianFlux1D, total_flux=totflux, mean=5000, fwhm=10)
        g2 = SourceSpectrum(
            GaussianFlux1D, total_flux=totflux, mean=6500, fwhm=100)
        g3 = SourceSpectrum(
            GaussianFlux1D, total_flux=totflux, mean=7500, fwhm=5)
        sp = SpectralElement(Box1D, x_0=1000, width=1) * (g1 + g2 + g3)
        assert_quantity_allclose(
            sp.waveset[::100],
            [999.49, 1000.49, 5019.95906231, 6699.59062307,
             7509.7672007] * u.AA)

    def test_redshift(self):
        tf_unit = u.erg / (u.cm * u.cm * u.s)
        sp = SourceSpectrum(
            GaussianFlux1D, total_flux=(1 * tf_unit), mean=5000, fwhm=10)
        sp.z = 1.3
        m = RedshiftScaleFactor(z=1.3)
        w_step25_z0 = [4978.76695499, 4989.3834775, 5000, 5010.6165225] * u.AA
        assert_quantity_allclose(sp.waveset[::25], m(w_step25_z0))

    def test_redshift_none(self):
        sp = SourceSpectrum(Const1D, amplitude=1, z=1.3)
        assert sp.waveset is None

    def test_complicated_tree(self):
        """Throw everything in and insert redshift and scale in the middle."""

        # On one side, we have a composite bandpass.
        bp1 = SpectralElement(Const1D, amplitude=1.01)
        bp2 = SpectralElement(
            Empirical1D, points=[4999, 5000.001, 5030], lookup_table=[0, 1, 0])
        bp = bp1 * (0.8 * bp2)  # [4999, 5000.001, 5030]

        # On the other side, we have composite spectrum with
        # scale and redshift.
        sp1 = SourceSpectrum(
            Empirical1D, points=[5001, 5011, 5020], lookup_table=[0, 1, 0])
        sp2 = SourceSpectrum(
            Empirical1D, points=[5000, 5010, 5020], lookup_table=[0, 1, 0])
        sp3 = sp2 + (sp1 * 0.5)  # [5000, 5001, 5010, 5011, 5020]
        sp3.z = 0.01  # [5050, 5051.01, 5060.1, 5061.11, 5070.2]
        sp = sp1 + sp3  # [5001, 5011, 5020, 5050, 5051.01, 5060.1, 5061.11, 5070.2]  # noqa

        sp_final = sp * bp
        np.testing.assert_array_equal(
            sp_final.waveset.value,
            [4999, 5000.001, 5001, 5011, 5020, 5030, 5050, 5051.01,
             5060.1, 5061.11, 5070.2])

    def test_exceptions(self):
        with pytest.raises(exceptions.SynphotError):
            get_waveset('foo')


class TestMathOperators:
    """Test spectrum math operators."""
    def setup_class(self):
        self.sp_1 = SourceSpectrum(
            Empirical1D, points=[3999.9, 4000.0, 5000.0, 6000.0, 6000.1],
            lookup_table=[0, 3.5e-14, 4e-14, 4.5e-14, 0] * units.FLAM)
        self.sp_2 = SourceSpectrum(
            Empirical1D, points=_wave, lookup_table=_flux_jy,
            meta={'PHOTLAM': [9.7654e-3, 1.003896e-2, 9.78473e-3]})
        self.bp_1 = SpectralElement(
            Empirical1D, points=[399.99, 400.01, 500.0, 590.0, 600.1] * u.nm,
            lookup_table=[0, 0.1, 0.2, 0.3, 0])

    def test_source_add(self):
        """Compare with ASTROLIB PYSYNPHOT."""
        ans = self.sp_1 + self.sp_2
        assert_quantity_allclose(
            ans(ans.waveset),
            [0.00976521, 0.01681283, 0.01970276, 0.01998463, 0.0197387,
             0.01985257, 0.02337638, 0.00978454] * units.PHOTLAM, rtol=1e-4)

    def test_source_sub(self):
        """Compare with ASTROLIB PYSYNPHOT."""
        ans = self.sp_1 - self.sp_2
        assert_quantity_allclose(
            ans(ans.waveset),
            [-9.76520783e-03, -2.71758275e-03, 1.72346256e-04, -9.29051118e-05,
             1.69629843e-04, 2.83499328e-04, 3.80731187e-03,
             -9.78453651e-03] * units.PHOTLAM,
            rtol=1e-4)

    def test_source_addsub_circular(self):
        """sp = sp + sp - sp"""
        ans = self.sp_1 + self.sp_1 - self.sp_1
        assert_quantity_allclose(ans(ans.waveset), self.sp_1(ans.waveset))

    def test_source_addsub_exception(self):
        with pytest.raises(exceptions.IncompatibleSources):
            self.sp_1 + self.bp_1

    @pytest.mark.parametrize('x', [2, 2 * u.dimensionless_unscaled])
    def test_source_mul_div_scalar(self, x):
        w = self.sp_1.waveset

        ans1 = self.sp_1 * x
        assert_quantity_allclose(
            ans1(w),
            [0, 0.01409552, 0.02013646, 0.02718424, 0] * units.PHOTLAM,
            rtol=1e-6)

        # rmul does not work with Quantity
        if not isinstance(x, u.Quantity):
            ans2 = x * self.sp_1
            assert_quantity_allclose(ans1(w), ans2(w), rtol=0)

        ans3 = self.sp_1 / x
        assert_quantity_allclose(
            ans3(w),
            [0, 0.00352388, 0.00503411, 0.00679606, 0] * units.PHOTLAM,
            atol=1e-7 * units.PHOTLAM)

    def test_source_mul_div_spec(self):
        """Compare mul with ASTROLIB PYSYNPHOT. Also test bp * sp."""
        ans1 = self.sp_1 * self.bp_1
        ans2 = self.bp_1 * self.sp_1
        w = ans1.waveset[:-1]
        assert_quantity_allclose(
            ans1(w),
            [0, 3.52381254e-04, 7.04792712e-04, 2.01360717e-03, 3.97184014e-03,
             4.03718269e-05, 0] * units.PHOTLAM, rtol=1e-4)
        assert_quantity_allclose(ans1(w), ans2(w), rtol=0)

        ans3 = self.sp_1 / self.bp_1
        assert_quantity_allclose(
            ans3(w),
            [0, 0.14095528, 0.07048066, 0.05034117, 0.04413243, 4.57601236,
             0] * units.PHOTLAM)

        ans4 = self.sp_1 / self.sp_1
        assert_quantity_allclose(
            ans4([4000, 5000, 6000]), 1 * u.dimensionless_unscaled)

        # Dividing throughput by flux does not make sense.
        with pytest.raises(exceptions.IncompatibleSources):
            self.bp_1 / self.sp_1

    def test_source_mul_div_exceptions(self):
        """Only mul is tested but truediv uses the same validation."""
        with pytest.raises(exceptions.IncompatibleSources):
            self.sp_1 * self.sp_2

        with pytest.raises(exceptions.IncompatibleSources):
            self.sp_1 * [1, 2]
        with pytest.raises(exceptions.IncompatibleSources):
            self.sp_1 * (1 - 1j)

        with pytest.raises(exceptions.IncompatibleSources):
            self.sp_1 * u.Quantity([1, 2])
        with pytest.raises(exceptions.IncompatibleSources):
            self.sp_1 * u.Quantity(1 - 1j)
        with pytest.raises(exceptions.IncompatibleSources):
            self.sp_1 * (1 * u.AA)

    def test_bandpass_addsub(self):
        """Not supported."""
        with pytest.raises(NotImplementedError):
            self.bp_1 + self.bp_1
        with pytest.raises(NotImplementedError):
            self.bp_1 + 2.0
        with pytest.raises(NotImplementedError):
            self.bp_1 - self.bp_1
        with pytest.raises(NotImplementedError):
            self.bp_1 - 2.0

    @pytest.mark.parametrize('x', [2.0, 2.0 * u.dimensionless_unscaled])
    def test_bandpass_mul_div_scalar(self, x):
        w = self.bp_1.waveset

        ans1 = self.bp_1 * x
        assert_quantity_allclose(ans1(w), [0, 0.2, 0.4, 0.6, 0])

        # rmul does not work with Quantity
        if not isinstance(x, u.Quantity):
            ans2 = x * self.bp_1
            assert_quantity_allclose(ans1(w), ans2(w), rtol=0)

        ans3 = self.bp_1 / x
        assert_quantity_allclose(ans3(w), [0, 0.05, 0.1, 0.15, 0])

    def test_bandpass_mul_div_bandpass(self):
        ans1 = self.bp_1 * self.bp_1
        assert_quantity_allclose(
            ans1(ans1.waveset), [0, 0.01, 0.04, 0.09, 0])

        w = [4000.1, 5000, 5900]  # Avoid div by zero
        ans2 = self.bp_1 / self.bp_1
        assert_quantity_allclose(ans2(w), 1)

    def test_bandpass_mul_div_exceptions(self):
        """Only mul is tested but truediv uses the same validation."""
        class DummyObject:
            pass

        with pytest.raises(exceptions.IncompatibleSources):
            self.bp_1 * DummyObject()

        with pytest.raises(exceptions.IncompatibleSources):
            self.bp_1 * u.Quantity([1, 2])
        with pytest.raises(exceptions.IncompatibleSources):
            self.bp_1 * u.Quantity(1 - 1j)
        with pytest.raises(exceptions.IncompatibleSources):
            self.bp_1 * (1 * u.AA)

        with pytest.raises(exceptions.IncompatibleSources):
            self.bp_1 * [1, 2]
        with pytest.raises(exceptions.IncompatibleSources):
            self.bp_1 * (1 - 1j)


class TestWriteSpec:
    """Test spectrum to_fits() method."""
    def setup_class(self):
        self.sp = SourceSpectrum(
            Empirical1D, points=_wave, lookup_table=_flux_photlam,
            meta={'expr': 'Test source'})
        self.bp = SpectralElement(
            Empirical1D, points=_wave, lookup_table=np.ones(_wave.shape),
            meta={'expr': 'Test bandpass'})

    @pytest.mark.parametrize(
        ('is_sp', 'ext_hdr'),
        [(True, None),
         (True, {'foo': 'foo'}),
         (False, None),
         (False, {'foo': 'foo'})])
    def test_write(self, tmp_path, is_sp, ext_hdr):
        outfile = str(tmp_path / 'outspec.fits')

        if is_sp:
            sp1 = self.sp
        else:
            sp1 = self.bp

        if ext_hdr is None:
            sp1.to_fits(outfile, overwrite=True, trim_zero=False,
                        pad_zero_ends=False)
        else:
            sp1.to_fits(outfile, overwrite=True, trim_zero=False,
                        pad_zero_ends=False, ext_header=ext_hdr)

        # Read it back in and check
        sp2 = sp1.__class__.from_file(outfile)
        np.testing.assert_allclose(sp2(sp2.waveset), sp1(sp1.waveset))
        hdr = fits.getheader(outfile, 1)
        assert 'expr' in hdr
        if ext_hdr is not None:
            assert 'foo' in hdr



# Licensed under a 3-clause BSD style license - see LICENSE.rst
"""Spectrum models not in `astropy.modeling`."""

# STDLIB
import math
import warnings
from copy import deepcopy
from functools import partial

# THIRD-PARTY
import numpy as np

# ASTROPY
from astropy import constants as const
from astropy import units as u
from astropy.modeling import Fittable1DModel, Model, Parameter
from astropy.modeling import models as _models
from astropy.modeling.models import (RickerWavelet1D as _RickerWavelet1D,
                                     Tabular1D)
from astropy.stats.funcs import gaussian_fwhm_to_sigma, gaussian_sigma_to_fwhm
from astropy.utils import metadata
from astropy.utils.exceptions import AstropyUserWarning

# LOCAL
from synphot import units
from synphot.exceptions import SynphotError
from synphot.utils import merge_wavelengths

__all__ = ['BlackBody1D', 'BlackBodyNorm1D', 'Box1D', 'ConstFlux1D',
           'Empirical1D', 'Gaussian1D', 'GaussianAbsorption1D',
           'GaussianFlux1D', 'Lorentz1D', 'MexicanHat1D', 'RickerWavelet1D',
           'PowerLawFlux1D', 'Trapezoid1D', 'get_waveset', 'get_metadata']


class BlackBody1D(Fittable1DModel):
    """Create a :ref:`blackbody spectrum <synphot-planck-law>`
    model with given temperature.

    Parameters
    ----------
    temperature : float
        Blackbody temperature in Kelvin.

    """
    temperature = Parameter(default=5000)

    def __init__(self, *args, **kwargs):
        super(BlackBody1D, self).__init__(*args, **kwargs)
        self.meta['expr'] = 'bb({0})'.format(self.temperature.value)

    @property
    def lambda_max(self):
        """Peak wavelength in Angstrom when the curve is expressed as
        power density."""
        return ((const.b_wien.value / self.temperature) * u.m).to_value(u.AA)

    def bounding_box(self, factor=10.0):
        """Tuple defining the default ``bounding_box`` limits,
        ``(x_low, x_high)``.

        .. math::

            x_{\\mathrm{low}} = 0

            x_{\\mathrm{high}} = \\log(\\lambda_{\\mathrm{max}} \\;\
            (1 + \\mathrm{factor}))

        Parameters
        ----------
        factor : float
            Used to calculate ``x_high``.

        """
        w0 = self.lambda_max
        return (w0 * 0, np.log10(w0 + factor * w0))

    def sampleset(self, factor_bbox=10.0, num=1000):
        """Return ``x`` array that samples the feature.

        Parameters
        ----------
        factor_bbox : float
            Factor for ``bounding_box`` calculations.

        num : int
            Number of points to generate.

        """
        w1, w2 = self.bounding_box(factor=factor_bbox)

        if self._n_models == 1:
            w = np.logspace(w1, w2, num)
        else:
            w = list(map(partial(np.logspace, num=num), w1, w2))

        return np.asarray(w)

    @staticmethod
    def evaluate(x, temperature):
        """Evaluate the model.

        Parameters
        ----------
        x : number or ndarray
            Wavelengths in Angstrom.

        temperature : number
            Temperature in Kelvin.

        Returns
        -------
        y : number or ndarray
            Blackbody radiation in PHOTLAM per steradian.

        """
        from synphot.blackbody import blackbody_nu

        # Silence Numpy
        old_np_err_cfg = np.seterr(all='ignore')

        wave = np.ascontiguousarray(x) * u.AA
        bbnu_flux = blackbody_nu(wave, temperature)
        bbflux = (bbnu_flux * u.sr).to(
            units.PHOTLAM, u.spectral_density(wave)) / u.sr  # PHOTLAM/sr

        # Restore Numpy settings
        np.seterr(**old_np_err_cfg)

        return bbflux.value

    def integrate(self, *args):
        with u.add_enabled_equivalencies(u.temperature()):
            t = u.Quantity(self.temperature, u.K)

        return (const.sigma_sb * t ** 4 / math.pi)  # per steradian


class BlackBodyNorm1D(BlackBody1D):
    """Create a normalized :ref:`blackbody spectrum <synphot-planck-law>`
    with given temperature.

    It is normalized by multiplying `BlackBody1D` result with a solid angle,
    :math:`\\Omega`, as defined below, where :math:`d` is 1 kpc:

    .. math::

        \\Omega = \\frac{\\pi R_{\\mathrm{Sun}}^{2}}{d^{2}}

    Parameters
    ----------
    temperature : float
        Blackbody temperature in Kelvin.

    """
    def __init__(self, *args, **kwargs):
        super(BlackBodyNorm1D, self).__init__(*args, **kwargs)
        self._omega = np.pi * (const.R_sun / const.kpc).value ** 2  # steradian

    def evaluate(self, x, temperature):
        """Evaluate the model.

        Parameters
        ----------
        x : number or ndarray
            Wavelengths in Angstrom.

        temperature : number
            Temperature in Kelvin.

        Returns
        -------
        y : number or ndarray
            Blackbody radiation in PHOTLAM.

        """
        bbflux = super(BlackBodyNorm1D, self).evaluate(x, temperature)
        return bbflux * self._omega

    def integrate(self, *args):
        return super().integrate(*args) * self._omega


class Box1D(_models.Box1D):
    """Same as `astropy.modeling.functional_models.Box1D`, except with
    ``sampleset`` defined.

    Parameters
    ----------
    step : float
        Distance of first and last points w.r.t. bounding box. In default units
        (nominally Angstrom). Defaults to 0.01

    """
    def __init__(self, *args, **kwargs):

        if "step" in kwargs:
            self.step = kwargs.pop("step")
        else:
            self.step = 0.01

        super(Box1D, self).__init__(*args, **kwargs)

    @staticmethod
    def _calc_sampleset(w1, w2, step, minimal):
        """Calculate sampleset for each model."""
        if minimal:
            arr = [w1 - step, w1, w2, w2 + step]
        else:
            arr = np.arange(w1 - step, w2 + step + step, step)

        return arr

    def sampleset(self, step=None, minimal=False):
        """Return ``x`` array that samples the feature.

        Parameters
        ----------
        step : float, optional
            Distance of first and last points w.r.t. bounding box. If None,
            set to attribute ``step``.

        minimal : bool
            Only return the minimal points needed to define the box;
            i.e., box edges and a point outside on each side.

        """
        if step is None:
            step = self.step

        w1, w2 = tuple(self.bounding_box.bounding_box())

        if self._n_models == 1:
            w = self._calc_sampleset(w1, w2, step, minimal)
        else:
            w = list(map(partial(
                self._calc_sampleset, step=step, minimal=minimal), w1, w2))

        return np.asarray(w)

    def integrate(self, *args):
        # TODO: Remove unit hardcoding when we use model with units natively.
        with u.add_enabled_equivalencies(u.spectral()):
            w = u.Quantity(self.width, u.AA)

        return self.amplitude * w


class ConstFlux1D(_models.Const1D):
    """One dimensional constant flux model.

    Flux that is constant in a given unit might not be constant in
    another unit. During evaluation, flux is always converted to PHOTLAM.

    For multiple ``n_models``, this model only accepts amplitudes of the
    same flux unit; e.g., ``[1, 2]`` or ``Quantity([1, 2], 'photlam')``.

    Parameters
    ----------
    amplitude : number or `~astropy.units.quantity.Quantity`
        Value and unit of the constant function.
        If not Quantity, assume the unit of PHOTLAM.

    """
    def __init__(self, amplitude, **kwargs):
        if not isinstance(amplitude, u.Quantity):
            amplitude = amplitude * units.PHOTLAM

        if amplitude.unit == u.STmag:
            a = units.convert_flux(1, amplitude, units.FLAM)
        elif amplitude.unit == u.ABmag:
            a = units.convert_flux(1, amplitude, units.FNU)
        elif (amplitude.unit.physical_type in
              ('spectral flux density', 'spectral flux density wav',
               'photon flux density', 'photon flux density wav')):
            a = amplitude
        else:
            raise NotImplementedError(
                '{0} not supported.'.format(amplitude.unit))

        self._flux_unit = a.unit
        super(ConstFlux1D, self).__init__(amplitude=a.value, **kwargs)

    def evaluate(self, x, *args):
        """One dimensional constant flux model function.

        Parameters
        ----------
        x : number or ndarray
            Wavelengths in Angstrom.

        Returns
        -------
        y : number or ndarray
            Flux in PHOTLAM.

        """
        a = (self.amplitude * np.ones_like(x)) * self._flux_unit
        y = units.convert_flux(x, a, units.PHOTLAM)
        return y.value

    def integrate(self, x):
        # TODO: Remove unit hardcoding when we use model with units natively.
        # TODO: We do not handle wav_unit as wave number nor energy for now.
        if any(['wav' in t for t in self._flux_unit.physical_type]):
            wav_unit = u.AA
        else:
            wav_unit = u.Hz
        with u.add_enabled_equivalencies(u.spectral()):
            x = u.Quantity(x, wav_unit)
            amp = u.Quantity(self.amplitude, self._flux_unit)

        return (max(x) - min(x)) * amp


class Empirical1D(Tabular1D):
    """Empirical (sampled) spectrum or bandpass model.

    .. note::

        This model requires `SciPy <https://www.scipy.org>`_ 0.14
        or later to be installed.

    Parameters
    ----------
    keep_neg : bool
        Convert negative ``lookup_table`` values to zeroes?
        This is to be consistent with ASTROLIB PYSYNPHOT.

    kwargs : dict
        Keywords for `~astropy.modeling.tabular.Tabular1D` model
        creation or :func:`~scipy.interpolate.interpn`.
        When ``fill_value=np.nan`` is given, extrapolation is done
        based on nearest end points on each end; This is the default
        behavior.

    """
    def __init__(self, **kwargs):

        # Manually insert user metadata here to accomodate any warning
        # from self._process_neg_flux()
        meta = kwargs.pop('meta', {})
        self.meta = meta
        if 'warnings' not in self.meta:
            self.meta['warnings'] = {}

        x = kwargs['points']
        y = kwargs['lookup_table']

        # Points can only be ascending for interpn()
        if x[-1] < x[0]:
            x = x[::-1]
            y = y[::-1]
            kwargs['points'] = x

        # Handle negative flux
        keep_neg = kwargs.pop('keep_neg', False)
        self._keep_neg = keep_neg
        y = self._process_neg_flux(x, y)

        kwargs['lookup_table'] = y
        super(Empirical1D, self).__init__(**kwargs)

        # Set non-default interpolation default values.
        # For tapered model, just fill with zero;
        # Otherwise, extrapolate like ASTROLIB PYSYNPHOT.
        self.bounds_error = kwargs.get('bounds_error', False)
        if self.is_tapered():
            self.fill_value = kwargs.get('fill_value', 0)
        else:
            self.fill_value = kwargs.get('fill_value', np.nan)

    def _process_neg_flux(self, x, y):
        """Remove negative flux."""

        if self._keep_neg:  # Nothing to do
            return y

        old_y = None

        if np.isscalar(y):  # pragma: no cover
            if y < 0:
                n_neg = 1
                old_x = x
                old_y = y
                y = 0
        else:
            x = np.asarray(x)  # In case input is just pure list
            y = np.asarray(y)
            i = np.where(y < 0)
            n_neg = len(i[0])
            if n_neg > 0:
                old_x = x[i]
                old_y = y[i]
                y[i] = 0

        if old_y is not None:
            warn_str = ('{0} bin(s) contained negative flux or throughput'
                        '; it/they will be set to zero.'.format(n_neg))
            warn_str += '\n  points: {0}\n  lookup_table: {1}'.format(
                old_x, old_y)  # Extra info
            self.meta['warnings'].update({'NegativeFlux': warn_str})
            warnings.warn(warn_str, AstropyUserWarning)

        return y

    def is_tapered(self):
        return np.array_equal(
            self.lookup_table[::self.lookup_table.size - 1], [0, 0])

    def sampleset(self):
        """Return array that samples the feature."""
        return np.squeeze(self.points)

    def evaluate(self, inputs):
        """Evaluate the model.

        Parameters
        ----------
        inputs : number or ndarray
            Wavelengths in same unit as ``points``.

        Returns
        -------
        y : number or ndarray
            Flux or throughput in same unit as ``lookup_table``.

        """
        y = super(Empirical1D, self).evaluate(inputs)

        # Assume NaN at both ends need to be extrapolated based on
        # nearest end point.
        if self.fill_value is np.nan:
            # Cannot use sampleset() due to ExtinctionModel1D.
            x = np.squeeze(self.points)

            # np.squeeze may throw unit away.
            if (isinstance(self.points, tuple) and
                    isinstance(self.points[0], u.Quantity) and
                    not isinstance(x, u.Quantity)):
                x = x * self.points[0].unit

            if np.isscalar(y):  # pragma: no cover
                if inputs < x[0]:
                    y = self.lookup_table[0]
                elif inputs > x[-1]:
                    y = self.lookup_table[-1]
            else:
                y[inputs < x[0]] = self.lookup_table[0]
                y[inputs > x[-1]] = self.lookup_table[-1]

        return self._process_neg_flux(inputs, y)


class BaseGaussian1D(_models.Gaussian1D):
    """Same as `astropy.modeling.functional_models.BaseGaussian1D`, except with
    ``sampleset`` defined.

    """
    _sqrt_2_pi = math.sqrt(2 * math.pi)

    def sampleset(self, factor_step=0.1, **kwargs):
        """Return ``x`` array that samples the feature.

        Parameters
        ----------
        factor_step : float
            Factor for sample step calculation. The step is calculated
            using ``factor_step * self.stddev``.

        kwargs : dict
            Keyword(s) for ``bounding_box`` calculation.
            Default ``factor`` is set to 5 to be compatible with
            ASTROLIB PYSYNPHOT.

        """
        if 'factor' not in kwargs:
            kwargs['factor'] = 5.0

        w1, w2 = self.bounding_box(**kwargs)
        dw = factor_step * self.stddev

        if self._n_models == 1:
            w = np.arange(w1, w2, dw)
        else:
            w = list(map(np.arange, w1, w2, dw))

        return np.asarray(w)


class Gaussian1D(BaseGaussian1D):
    """Same as `astropy.modeling.functional_models.Gaussian1D`, except with
    ``sampleset`` defined.

    """
    def integrate(self, *args):
        # TODO: Remove unit hardcoding when we use model with units natively.
        with u.add_enabled_equivalencies(u.spectral()):
            stddev = u.Quantity(self.stddev, u.AA)

        return self.amplitude * stddev * self._sqrt_2_pi


# TODO: Deprecate this?
# This is not really supported anymore but kept for backward compatibility.
class GaussianAbsorption1D(BaseGaussian1D):
    """Same as ``astropy.modeling.functional_models.GaussianAbsorption1D``,
    except with ``sampleset`` defined.

    """
    @staticmethod
    def evaluate(x, amplitude, mean, stddev):
        """
        GaussianAbsorption1D model function.
        """
        return 1.0 - Gaussian1D.evaluate(x, amplitude, mean, stddev)

    @staticmethod
    def fit_deriv(x, amplitude, mean, stddev):
        """
        GaussianAbsorption1D model function derivatives.
        """
        import operator
        return list(map(
            operator.neg, Gaussian1D.fit_deriv(x, amplitude, mean, stddev)))


class GaussianFlux1D(Gaussian1D):
    """Same as `Gaussian1D` but accepts extra keywords below.

    Parameters
    ----------
    amplitude : float
        Amplitude of the Gaussian in PHOTLAM.
        Also see ``total_flux``.

    mean : float
        Mean of the Gaussian in Angstrom.

    stddev : float
        Standard deviation of the Gaussian in Angstrom.
        Also see ``fwhm``.

    fwhm : float
        Full width at half maximum of the Gaussian in Angstrom.
        If given, this overrides ``stddev``.

    total_flux : float
        Total flux under the Gaussian in ``erg/s/cm^2``.
        If given, this overrides ``amplitude``.

    """
    def __init__(self, *args, **kwargs):
        fwhm = kwargs.pop('fwhm', None)
        total_flux = kwargs.pop('total_flux', None)

        super(GaussianFlux1D, self).__init__(*args, **kwargs)

        if fwhm is None:
            fwhm = self.stddev * gaussian_sigma_to_fwhm
        else:
            self.stddev = fwhm * gaussian_fwhm_to_sigma

        gaussian_amp_to_totflux = self._sqrt_2_pi * self.stddev

        if total_flux is None:
            u_str = 'PHOTLAM'
            total_flux = self.amplitude * gaussian_amp_to_totflux
        else:
            u_str = 'FLAM'
            # total_flux is passed in unaltered, any conversion error would
            # happen here.
            tf_unit = u.erg / (u.cm * u.cm * u.s)
            if isinstance(total_flux, u.Quantity):
                total_flux = total_flux.to(tf_unit)
            else:
                total_flux = total_flux * tf_unit
            self.amplitude = (total_flux / (gaussian_amp_to_totflux * u.AA)).to_value(units.PHOTLAM, u.spectral_density(self.mean.value * u.AA))  # noqa
            total_flux = total_flux.value

        self.meta['expr'] = 'em({0:g}, {1:g}, {2:g}, {3})'.format(
            self.mean.value, fwhm, total_flux, u_str)

    def integrate(self, *args):
        # TODO: Remove unit hardcoding when we use model with units natively.
        return super(GaussianFlux1D, self).integrate(*args) * units.PHOTLAM


class Lorentz1D(_models.Lorentz1D):
    """Same as `astropy.modeling.functional_models.Lorentz1D`, except with
    ``sampleset`` defined.

    """
    def sampleset(self, factor_step=0.05, **kwargs):
        """Return ``x`` array that samples the feature.

        Parameters
        ----------
        factor_step : float
            Factor for sample step calculation. The step is calculated
            using ``factor_step * self.fwhm``.

        kwargs : dict
            Keyword(s) for ``bounding_box`` calculation.

        """
        w1, w2 = self.bounding_box(**kwargs)
        dw = factor_step * self.fwhm

        if self._n_models == 1:
            w = np.arange(w1, w2, dw)
        else:
            w = list(map(np.arange, w1, w2, dw))

        return np.asarray(w)

    def integrate(self, x):
        # TODO: Remove unit hardcoding when we use model with units natively.
        with u.add_enabled_equivalencies(u.spectral()):
            x = u.Quantity(x, u.AA)
            x_0 = u.Quantity(self.x_0, u.AA)
            gamma = u.Quantity(self.fwhm, u.AA) * 0.5

        a1 = np.arctan((min(x) - x_0) / gamma)
        a2 = np.arctan((max(x) - x_0) / gamma)
        da = (a2 - a1).to(u.dimensionless_unscaled, u.dimensionless_angles())
        return self.amplitude * gamma * da


class RickerWavelet1D(_RickerWavelet1D):
    """Same as `astropy.modeling.functional_models.RickerWavelet1D`, except
    with ``sampleset`` defined.

    """
    def sampleset(self, factor_step=0.1, **kwargs):
        """Return ``x`` array that samples the feature.

        Parameters
        ----------
        factor_step : float
            Factor for sample step calculation. The step is calculated
            using ``factor_step * self.sigma``.

        kwargs : dict
            Keyword(s) for ``bounding_box`` calculation.

        """
        w1, w2 = self.bounding_box(**kwargs)
        dw = factor_step * self.sigma

        if self._n_models == 1:
            w = np.arange(w1, w2, dw)
        else:
            w = list(map(np.arange, w1, w2, dw))

        return np.asarray(w)

    def integrate(self, x):
        # TODO: Remove unit hardcoding when we use model with units natively.
        with u.add_enabled_equivalencies(u.spectral()):
            x = u.Quantity(x, u.AA)
            x_0 = u.Quantity(self.x_0, u.AA)
            sig = u.Quantity(self.sigma, u.AA)

        # Roots, where y=0
        root_left = x_0 - sig
        root_right = x_0 + sig

        x_min = min(x)
        x_max = max(x)
        if x_min >= root_left or x_max <= root_right:
            raise NotImplementedError(
                'Partial analytic integration not supported')

        sig2 = sig * sig

        def _int_subregion(xx1, xx2):
            dx_min = xx1 - x_0
            dx_max = xx2 - x_0
            a1 = dx_min * np.exp(-0.5 * dx_min * dx_min / sig2)
            a2 = dx_max * np.exp(-0.5 * dx_max * dx_max / sig2)
            return abs(a2 - a1)

        # Unsigned area
        return self.amplitude * (_int_subregion(x_min, root_left) +
                                 _int_subregion(root_left, root_right) +
                                 _int_subregion(root_right, x_max))


# TODO: Emit proper deprecation warning.
# https://github.com/spacetelescope/synphot_refactor/issues/249
class MexicanHat1D(RickerWavelet1D):
    """This is the deprecated name for `RickerWavelet1D`."""


class PowerLawFlux1D(_models.PowerLaw1D):
    """One dimensional power law model with proper flux handling.

    For multiple ``n_models``, this model only accepts parameters of the
    same unit; e.g., ``amplitude=[1, 2]`` or
    ``amplitude=Quantity([1, 2], 'photlam')``.

    Also see `~astropy.modeling.powerlaws.PowerLaw1D`.

    Parameters
    ----------
    amplitude : number or `~astropy.units.quantity.Quantity`
        Model amplitude at the reference point.
        If not Quantity, assume the unit of PHOTLAM.

    x_0 : number or `~astropy.units.quantity.Quantity`
        Reference point.
        If not Quantity, assume the unit of Angstrom.

    alpha : float
        Power law index.

    """
    def __init__(self, amplitude, x_0, alpha, **kwargs):
        if not isinstance(amplitude, u.Quantity):
            amplitude = amplitude * units.PHOTLAM

        if (amplitude.unit.physical_type in
                ('spectral flux density', 'spectral flux density wav',
                 'photon flux density', 'photon flux density wav')):
            self._flux_unit = amplitude.unit
        else:
            raise NotImplementedError(
                '{0} not supported.'.format(amplitude.unit))

        if isinstance(x_0, u.Quantity):
            x_0 = x_0.to_value(u.AA, u.spectral())

        super(PowerLawFlux1D, self).__init__(
            amplitude=amplitude.value, x_0=x_0, alpha=alpha, **kwargs)

    def evaluate(self, x, *args):
        """Return flux in PHOTLAM. Assume input wavelength is in Angstrom."""
        xx = x / self.x_0
        y = (self.amplitude * xx ** (-self.alpha)) * self._flux_unit
        flux = units.convert_flux(x, y, units.PHOTLAM)
        return flux.value

    def integrate(self, x):
        # TODO: Remove unit hardcoding when we use model with units natively.
        with u.add_enabled_equivalencies(u.spectral()):
            x = u.Quantity(x, u.AA)
            x_0 = u.Quantity(self.x_0, u.AA)
            amp = u.Quantity(self.amplitude, self._flux_unit)

        fac = 1 - self.alpha
        denom = x_0 ** -self.alpha * fac
        return amp * (max(x) ** fac - min(x) ** fac) / denom


class Trapezoid1D(_models.Trapezoid1D):
    """Same as `astropy.modeling.functional_models.Trapezoid1D`, except with
    ``sampleset`` defined.

    """
    def sampleset(self):
        """Return ``x`` array that samples the feature."""
        x1, x4 = tuple(self.bounding_box.bounding_box())
        dw = self.width * 0.5
        x2 = self.x_0 - dw
        x3 = self.x_0 + dw

        if self._n_models == 1:
            w = [x1, x2, x3, x4]
        else:
            w = list(zip(x1, x2, x3, x4))

        return np.asarray(w)

    def integrate(self, *args):
        # TODO: Remove unit hardcoding when we use model with units natively.
        with u.add_enabled_equivalencies(u.spectral()):
            width = u.Quantity(self.width, u.AA)
            slope = u.Quantity(self.slope, 1 / u.AA)

        return self.amplitude * (width + self.amplitude / slope)


# Functions below are for sampleset magic.

def _get_sampleset(model):
    """Return sampleset of a model or `None` if undefined."""
    w = None
    if isinstance(model, Model) and hasattr(model, 'sampleset'):
        w = model.sampleset()
    return w


def _model_tree_evaluate_sampleset(root):
    # Not a CompoundModel, grab sampleset and be done.
    if not hasattr(root, 'op'):
        return _get_sampleset(root)

    model1 = root.left
    model2 = root.right

    # model2 is redshifted, apply the redshift if applicable.
    if isinstance(model1, _models.RedshiftScaleFactor):
        val = _model_tree_evaluate_sampleset(model2)
        if val is None:
            w = val
        else:
            w = model1.inverse(val)

    # This should not ever happen, so ignore the redshift.
    elif isinstance(model2, _models.RedshiftScaleFactor):
        w = _model_tree_evaluate_sampleset(model1)

    # One of the models is scaled. Non-redshift scaling does
    # not affect sampleset of the model.
    elif isinstance(model1, _models.Scale):
        w = _model_tree_evaluate_sampleset(model2)
    elif isinstance(model2, _models.Scale):
        w = _model_tree_evaluate_sampleset(model1)

    # Combine sampleset from both models.
    else:
        w1 = _model_tree_evaluate_sampleset(model1)
        w2 = _model_tree_evaluate_sampleset(model2)
        w = merge_wavelengths(w1, w2)

    return w


def get_waveset(model):
    """Get optimal wavelengths for sampling a given model.

    Parameters
    ----------
    model : `~astropy.modeling.Model`
        Model.

    Returns
    -------
    waveset : array-like or `None`
        Optimal wavelengths. `None` if undefined.

    Raises
    ------
    synphot.exceptions.SynphotError
        Invalid model.

    """
    if not isinstance(model, Model):
        raise SynphotError('{0} is not a model.'.format(model))

    return _model_tree_evaluate_sampleset(model)


# Functions below are for meta magic.

def _get_meta(model):
    """Return metadata of a model."""
    w = {}
    if isinstance(model, Model):
        w = model.meta
    return w


def _model_tree_evaluate_metadata(root):
    # Not a CompoundModel, grab metadata and be done.
    if not hasattr(root, 'op'):
        return _get_meta(root)

    m1 = _model_tree_evaluate_metadata(root.left)
    m2 = _model_tree_evaluate_metadata(root.right)
    return metadata.merge(m1, m2, metadata_conflicts='silent')


def get_metadata(model):
    """Get metadata for a given model.

    Parameters
    ----------
    model : `~astropy.modeling.Model`
        Model.

    Returns
    -------
    meta : dict
        Metadata for the model.

    Raises
    ------
    synphot.exceptions.SynphotError
        Invalid model.

    """
    if not isinstance(model, Model):
        raise SynphotError('{0} is not a model.'.format(model))

    # Deep copy to make sure modiyfing returned metadata
    # does not modify input model metadata, especially
    # if input model is not a compound model.
    meta = deepcopy(_model_tree_evaluate_metadata(model))

    return meta
